from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3


def define_pour_points_with_lakes(
    grass,
    con,
    garray,
    remove_lake_inlets = False,
    str_r="str_grass_r",
    cat_no_lake="cat_no_lake",
    sl_lakes="sl_lakes",
    sl_connected_lake="sl_connected_lake",
    sl_str_connected_lake="sl_str_connected_lake",
    acc="acc",
    pourpoints_with_lakes="pourpoints_with_lakes",
    lake_inflow_pourpoints="lake_inflow_pourpoints",
    lake_outflow_pourpoints="lake_outflow_pourpoints",
    catchment_pourpoints_outside_lake="catchment_pourpoints_outside_lake",
):
    r_nuill_n = 5000
    # define catchment pourpoints and routing info
    routing_info = generate_routing_info_of_catchments(
        grass, con, cat=cat_no_lake, acc=acc, str=str_r, Name="cat1",garray=garray
    )
    routing_info = routing_info.fillna(-1)

    # define lake outlets
    # lake maximum acc 
    grass.run_command(
        "r.stats.zonal",
        base=sl_lakes,
        cover=acc,
        method="max",
        output="sl_lakes_maxacc",
        overwrite=True,
    )
    # get lake output 
    exp = "%s =if( abs(%s - %s) < 1.0e-15,%s,null())" % (
        lake_outflow_pourpoints+'t',
        acc,
        "sl_lakes_maxacc",
        sl_lakes,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    ##################################################################
    # check if got mult points
    # only keep 1 points
    grass.run_command(
        "r.to.vect",
        input=lake_outflow_pourpoints+'t',
        output=lake_outflow_pourpoints+'t',
        type="point",
#        flags="v",
        overwrite=True,
    )
    grass_raster_v_to_raster(
        grass,
        input=lake_outflow_pourpoints+'t',
        output=lake_outflow_pourpoints+'t2',
        column="#",
        use="cat",
    )

    grass.run_command("v.db.addcolumn", map=lake_outflow_pourpoints+'t', columns="lakeid int")
    grass.run_command("v.what.rast", map=lake_outflow_pourpoints+'t', raster=acc, column="OL_acc")
    grass.run_command("v.what.rast", map=lake_outflow_pourpoints+'t', raster=lake_outflow_pourpoints+'t', column="lakeid")
    sqlstat = "SELECT cat, OL_acc,lakeid FROM %s" % (
        lake_outflow_pourpoints+'t'
    )
    outletinfo_temp = pd.read_sql_query(sqlstat, con)
    outletinfo_temp['count'] = outletinfo_temp.groupby('lakeid')['lakeid'].transform('count')
    outletinfo_temp['maxacc'] = outletinfo_temp.groupby('lakeid')['OL_acc'].transform('max')
    outletinfo_temp = outletinfo_temp.sort_values(by='OL_acc', ascending=False)
    outletinfo_temp = outletinfo_temp.drop_duplicates(subset=['lakeid'], keep='first')
    extract_lake = outletinfo_temp['cat'].values

    array_lake_OL2 = garray.array(mapname=lake_outflow_pourpoints+'t2')
    mask = np.logical_not(np.isin(array_lake_OL2, extract_lake))
    array_lake_OL2[mask] = -9999
    temparray = garray.array()
    temparray[:, :] = array_lake_OL2[:, :]
    temparray.write(mapname=lake_outflow_pourpoints + "t3", overwrite=True)
    grass.run_command("r.null", map=lake_outflow_pourpoints + "t3", setnull=[-9999])
    exp = "%s = int(%s)" % (
        lake_outflow_pourpoints + "t3",
        lake_outflow_pourpoints + "t3",
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    exp = "%s = if(isnull(%s),null(),%s)" % (
        lake_outflow_pourpoints,
        lake_outflow_pourpoints + "t3",
        sl_lakes,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    ##################################################################


    # add lake inlets

    #### overlay str and lakes
    exp = "%s = if(isnull(int(%s)),null(),%s)" % (
        "connect_lake_str",
        sl_str_connected_lake,
        str_r,
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    #### create a unique id for overlaied lake and river
    grass.run_command(
        "r.cross",
        input=["connect_lake_str", sl_str_connected_lake],
        output="unique_lake_str",
        flags="z",
        overwrite=True,
    )
        
    ##### obtain lake id and coresponding overlaied str id
    riv_lake_id, str_id = generate_stats_list_from_grass_raster(
        grass, mode=2, input_a="unique_lake_str", input_b="connect_lake_str"
    )

    riv_lake_id2, cl_id = generate_stats_list_from_grass_raster(
        grass, mode=2, input_a="unique_lake_str", input_b=sl_str_connected_lake
    )

    ### get the lake raster value at the begining of each river segments,if the river
    ### the beginning of the river segment located within the lake, than it can not
    ### be used as the lake inflow segment.
    grass.run_command(
        "v.what.rast", map="cat1_IL_v", raster=sl_connected_lake, column="sl_cl_id"
    )
    sqlstat = "SELECT IL_SubId, sl_cl_id FROM %s" % ("cat1_IL_v")
    str_start_pt_lakeid = pd.read_sql_query(sqlstat, con)
    str_start_pt_lakeid = str_start_pt_lakeid.fillna(-1)

    (
        str_id_within_lakes,
        non_lake_inflow_segs,
        str_id_lake_inlfow,
    ) = return_non_lake_inflow_segs_and_segs_within_lakes(
        riv_lake_id, str_id, riv_lake_id2, cl_id, routing_info, str_start_pt_lakeid
    )

    # remove cat outlet that within the lake
    
    grass_raster_setnull_array(input = "cat1_OL",output = catchment_pourpoints_outside_lake,values = str_id_within_lakes,grass = grass)
    grass_raster_setnull_array(input = "unique_lake_str",output = "unique_lake_str_inflow",values = non_lake_inflow_segs, grass = grass)
    
    # if len(str_id_within_lakes) < r_nuill_n:
    #     grass.run_command(
    #         "g.copy", rast=("cat1_OL", catchment_pourpoints_outside_lake), overwrite=True
    #     )
    #     grass.run_command(
    #         "r.null",
    #         map=catchment_pourpoints_outside_lake,
    #         setnull=str_id_within_lakes,
    #         overwrite=True,
    #     )
    # else:
    #     grass.run_command(
    #         "g.copy", rast=("cat1_OL", catchment_pourpoints_outside_lake), overwrite=True
    #     )
    #     k=0
    #     for i in range(0,int(len(str_id_within_lakes)/r_nuill_n) + 1):
    #         print("r.null   ",k,i*r_nuill_n+r_nuill_n,len(str_id_within_lakes))
    #         if i*r_nuill_n+r_nuill_n < len(str_id_within_lakes):
    #             grass.run_command(
    #                 "r.null",
    #                 map=catchment_pourpoints_outside_lake,
    #                 setnull=str_id_within_lakes[k:i*r_nuill_n+r_nuill_n],
    #                 overwrite=True,
    #             ) 
    #         elif  k > len(str_id_within_lakes):
    #             break
    #         else:
    #             grass.run_command(
    #                 "r.null",
    #                 map=catchment_pourpoints_outside_lake,
    #                 setnull=str_id_within_lakes[k:len(str_id_within_lakes) - 1],
    #                 overwrite=True,
    #             )      
    #         k = i*r_nuill_n+r_nuill_n                             
        # catchment_pourpoints_outside_lake_array = garray.array(mapname="cat1_OL")
        # mask = np.isin(catchment_pourpoints_outside_lake_array, str_id_within_lakes)
        # catchment_pourpoints_outside_lake_array[mask] = -9999
        # temparray = garray.array()
        # temparray[:, :] = catchment_pourpoints_outside_lake_array[:, :]
        # temparray.write(mapname=catchment_pourpoints_outside_lake, overwrite=True)
        # grass.run_command("r.null", map=catchment_pourpoints_outside_lake, setnull=[-9999])
        # exp = "%s = int(%s)" % (
        #     catchment_pourpoints_outside_lake,
        #     catchment_pourpoints_outside_lake,
        # )
        # grass.run_command("r.mapcalc", expression=exp, overwrite=True)
                
    # remove non lake inflow river segment
    # if len(non_lake_inflow_segs) < r_nuill_n:
    #     grass.run_command(
    #         "g.copy", rast=("unique_lake_str", "unique_lake_str_inflow"), overwrite=True
    #     )
    #     grass.run_command(
    #         "r.null",
    #         map="unique_lake_str_inflow",
    #         setnull=non_lake_inflow_segs,
    #         overwrite=True,
    #     )
    # else:
    #     grass.run_command(
    #         "g.copy", rast=("unique_lake_str", "unique_lake_str_inflow"), overwrite=True
    #     )
    #     k=0
    #     for i in range(0,int(len(non_lake_inflow_segs)/r_nuill_n)+1):
    #         print("r.null   ",k,i*r_nuill_n+r_nuill_n,len(non_lake_inflow_segs))
    #         if i*r_nuill_n+r_nuill_n < len(non_lake_inflow_segs):
    #             grass.run_command(
    #                 "r.null",
    #                 map="unique_lake_str_inflow",
    #                 setnull=non_lake_inflow_segs[k:i*r_nuill_n+r_nuill_n],
    #                 overwrite=True,
    #             )
    #         elif  k > len(non_lake_inflow_segs):
    #             break
    #         else:
    #             grass.run_command(
    #                 "r.null",
    #                 map="unique_lake_str_inflow",
    #                 setnull=non_lake_inflow_segs[k:len(non_lake_inflow_segs) - 1],
    #                 overwrite=True,
    #             )      
    #         k = i*r_nuill_n+r_nuill_n  
    
    ## find lake inflow points from inflow segments

    # get all str that are lake inflow str
    all_river_ids, temp = generate_stats_list_from_grass_raster(
        grass, mode=1, input_a=str_r
    )
    str_id_non_lake_inlfow = [x for x in all_river_ids if x not in str_id_lake_inlfow]
    
    grass_raster_setnull_array(input = str_r,output = "lake_inflow_str",values = str_id_non_lake_inlfow,grass = grass)
    
    # if len(str_id_non_lake_inlfow) < r_nuill_n:
    #     grass.run_command("g.copy", rast=(str_r, "lake_inflow_str"), overwrite=True)
    #     grass.run_command(
    #         "r.null", map="lake_inflow_str", setnull=str_id_non_lake_inlfow, overwrite=True
    #     )
    # else:
    #     grass.run_command("g.copy", rast=(str_r, "lake_inflow_str"), overwrite=True)
    #     k=0
    #     for i in range(0,int(len(str_id_non_lake_inlfow)/r_nuill_n)+1):
    #         print("r.null   ",k,i*r_nuill_n+r_nuill_n,len(str_id_non_lake_inlfow))
    #         if i*r_nuill_n+r_nuill_n < len(str_id_non_lake_inlfow):
    #             grass.run_command(
    #                 "r.null",
    #                 map="lake_inflow_str",
    #                 setnull=str_id_non_lake_inlfow[k:i*r_nuill_n+r_nuill_n],
    #                 overwrite=True,
    #             )
    #         elif  k > len(str_id_non_lake_inlfow):
    #             break
    #         else:
    #             grass.run_command(
    #                 "r.null",
    #                 map="lake_inflow_str",
    #                 setnull=str_id_non_lake_inlfow[k:len(str_id_non_lake_inlfow) - 1],
    #                 overwrite=True,
    #             )      
    #         k = i*r_nuill_n+r_nuill_n  
            
        # lake_inflow_str_array = garray.array(mapname=str_r)
        # mask = np.isin(lake_inflow_str_array, str_id_non_lake_inlfow)
        # lake_inflow_str_array[mask] = -9999
        # temparray = garray.array()
        # temparray[:, :] = lake_inflow_str_array[:, :]
        # temparray.write(mapname="lake_inflow_str", overwrite=True)
        # grass.run_command("r.null", map="lake_inflow_str", setnull=[-9999])
        # exp = "%s = int(%s)" % (
        #     "lake_inflow_str",
        #     "lake_inflow_str",
        # )
        # grass.run_command("r.mapcalc", expression=exp, overwrite=True)
            
    # find the lake inlet point within the lake

    grass.run_command(
        "r.stats.zonal",
        base="unique_lake_str_inflow",
        cover="cat1_acc_riv",
        method="min",
        output="lake_inflow_seg_minacc",
        overwrite=True,
    )
    ### Find the grid that equal to the max acc, thus this is the outlet grids

    exp = "%s =if( abs(%s - %s) < 1.0e-15,%s,null())" % (
        "lake_inflow_IL",
        "cat1_acc_riv",
        "lake_inflow_seg_minacc",
        "unique_lake_str_inflow",
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    # extent the inflow seg to lake outside
    grass.run_command(
        "r.grow",
        input="lake_inflow_IL",
        output="lake_inflow_IL_grow2",
        radius=1.5,
        overwrite=True,
    )

    # set grids in lakes to null
    exp = "%s = if(isnull(%s),int(%s),null())" % (
        "lake_inflow_IL_grow",
        sl_connected_lake,
        "lake_inflow_IL_grow2",
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)


    # set non river cells to non
    exp = "%s = if(isnull('%s'),null(),int(%s))" % (
        "extented_lake_inflow_seg",
        "lake_inflow_str",
        "lake_inflow_IL_grow",
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    grass.run_command(
        "r.stats.zonal",
        base="extented_lake_inflow_seg",
        cover="cat1_acc_riv",
        method="min",
        output="extented_lake_inflow_seg_minacc",
        overwrite=True,
    )
    ### Find the grid that equal to the max acc, thus this is the outlet grids

    exp = "%s =if(abs(%s - %s) < 1.0e-15,%s,null())" % (
        lake_inflow_pourpoints,
        "cat1_acc_riv",
        "extented_lake_inflow_seg_minacc",
        "extented_lake_inflow_seg",
    )
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    if remove_lake_inlets:
        exp = "%s =if(int(%s) > -1000000,null(),null())" % (
            lake_inflow_pourpoints+'t',
            lake_inflow_pourpoints,
        )
        grass.run_command("r.mapcalc", expression=exp, overwrite=True)

        exp = "%s = int(%s)" % (
            lake_inflow_pourpoints,
            lake_inflow_pourpoints+'t',
        )

        grass.run_command("r.mapcalc", expression=exp, overwrite=True)


    #### create a unique id for overlaied lake and river
    grass.run_command(
        "r.cross",
        input=[
            lake_inflow_pourpoints,
            catchment_pourpoints_outside_lake,
            lake_outflow_pourpoints,
        ],
        output=pourpoints_with_lakes,
        overwrite=True,
    )

    exp = "%s = %s + 1" % (pourpoints_with_lakes, pourpoints_with_lakes)
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    grass.run_command(
        "r.to.vect",
        input=pourpoints_with_lakes,
        output=pourpoints_with_lakes,
        type="point",
        overwrite=True,
        flags="v",
    )

    grass.run_command(
        "r.to.vect",
        input=lake_inflow_pourpoints,
        output=lake_inflow_pourpoints,
        type="point",
        overwrite=True,
        flags="v",
    )
    grass.run_command(
        "r.to.vect",
        input=catchment_pourpoints_outside_lake,
        output=catchment_pourpoints_outside_lake,
        type="point",
        overwrite=True,
        flags="v",
    )
    grass.run_command(
        "r.to.vect",
        input=lake_outflow_pourpoints,
        output=lake_outflow_pourpoints,
        type="point",
        overwrite=True,
        flags="v",
    )

    #
    ##### obtain lake id and coresponding overlaied str id
    CL_Id, Str_Id = generate_stats_list_from_grass_raster(
        grass, mode=2, input_a=sl_connected_lake, input_b=str_r
    )

    Lakes_WIth_Multi_Outlet, Remove_Str = Check_If_Lake_Have_Multi_OutLet(
        CL_Id, Str_Id, routing_info
    )

    un_used_map = [
                   "sl_lakes_maxacc",
                   lake_outflow_pourpoints+'t',
                   lake_outflow_pourpoints+'t2',
                   lake_outflow_pourpoints + 't3',
                   "connect_lake_str",
                   "unique_lake_str_inflow",
                   "lake_inflow_str",
                   "lake_inflow_seg_minacc",
                   "lake_inflow_IL",
                   "lake_inflow_IL_grow2",
                   "lake_inflow_IL_grow",
                   "extented_lake_inflow_seg",
                   "extented_lake_inflow_seg_minacc",
    ]
    grass.run_command(
        "g.remove",
        type="raster",
        name=un_used_map,
        flags="f",
    )
    
    return Lakes_WIth_Multi_Outlet, Remove_Str



from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import pandas as pd
import numpy as np
import shutil
import tempfile


def simplify_routing_structure_by_filter_lakes_qgis(
    Routing_Product_Folder = '#',
    Thres_Area_Conn_Lakes=-1,
    Thres_Area_Non_Conn_Lakes=-1,
    Selected_Lake_List_in=[],
    OutputFolder="#",
    qgis_prefix_path="#",
    gis_platform="qgis",
):
    """Simplify the routing product by lake area

    Function that used to simplify the routing product by user
    provided lake area thresthold.
    The input catchment polygons is the routing product before
    merging for lakes. It is provided with the routing product.
    The result is the simplified catchment polygons. But
    result from this fuction still not merging catchment
    covering by the same lake. Thus, The result generated
    from this tools need further processed by
    Define_Final_Catchment, or can be further processed by
    Customize_Routing_Topology

    Parameters
    ----------

    Path_final_riv_ply             : string
        Path to the catchment polygon which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.
    Path_final_riv                 : string
        Path to the river polyline which is the routing product
        before merging lakes catchments and need to be processed before
        used. It is the input for simplify the routing product based
        on lake area or drianage area.
    Path_Con_Lake_ply              : string
        Path to a connected lake polygon. Connected lakes are lakes that
        are connected by Path_final_riv.
    Path_NonCon_Lake_ply           : string
        Path to a non connected lake polygon. Connected lakes are lakes
        that are not connected by Path_final_riv.
    Thres_Area_Conn_Lakes          : float (optional)
        It is the lake area threshold for connated lakes, in km2
    Thres_Area_Non_Conn_Lakes      : float (optional)
        It is the lake area threshold for non connated lakes, in km2
    Selection_Method               : string
        It is a string indicate lake selection methods
        "ByArea" means lake in the routing product will be selected based
        on two lake area thresthold Thres_Area_Conn_Lakes and
        Thres_Area_Non_Conn_Lakes
        "ByLakelist" means lake in the routing product will be selected
        based on user provided hydrolake id, in Selected_Lake_List_in
    Selected_Lake_List_in          : list
        A list of lake ids that will be keeped in the routing product.
        Lakes not in the list will be removed from routing product.
    OutputFolder                   : string
        Folder name that stores generated extracted routing product

    Notes
    -------
    This function has no return values, instead will generate following
    files. The output tpye will be the same as inputs, but the routing
    network will be simplified by removing lakes.

    os.path.join(OutputFolder,os.path.basename(Path_final_riv_ply))
    os.path.join(OutputFolder,os.path.basename(Path_final_riv))
    os.path.join(OutputFolder,os.path.basename(Path_Con_Lake_ply))
    os.path.join(OutputFolder,os.path.basename(Path_NonCon_Lake_ply))

    Returns:
    -------
    None

    Examples
    -------

    """
    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    
    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)

    tempfolder = os.path.join(
        tempfile.gettempdir(),
        "basinmaker_sllake_" + str(np.random.randint(1, 10000 + 1)),
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)


    Path_Catchment_Polygon="#"
    Path_River_Polyline="#"
    Path_Con_Lake_ply="#"
    Path_NonCon_Lake_ply="#"
    Path_obs_gauge_point="#"
    Path_final_cat_ply="#"
    Path_final_cat_riv="#"
    ##define input files from routing prodcut 
    for file in os.listdir(Routing_Product_Folder):
        if file.endswith(".shp"):
            if 'catchment_without_merging_lakes' in file:
                Path_Catchment_Polygon = os.path.join(Routing_Product_Folder, file)
            if 'river_without_merging_lakes' in file:
                Path_River_Polyline = os.path.join(Routing_Product_Folder, file)
            if 'sl_connected_lake' in file:
                Path_Con_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'sl_non_connected_lake' in file:
                Path_NonCon_Lake_ply = os.path.join(Routing_Product_Folder, file)
            if 'obs_gauges' in file:
                Path_obs_gauge_point = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info' in file:
                Path_final_cat_ply = os.path.join(Routing_Product_Folder, file)
            if 'finalcat_info_riv' in file:
                Path_final_cat_riv = os.path.join(Routing_Product_Folder, file)                

    if Path_Catchment_Polygon == '#' or  Path_River_Polyline =='#':
        print("Invalid routing product folder ")

    Path_final_riv_ply = Path_Catchment_Polygon
    Path_final_riv = Path_River_Polyline
    
    ## copy obs_gauges to output folder 
    for file in os.listdir(Routing_Product_Folder):
        if 'obs_gauges' in file:
            shutil.copy(os.path.join(Routing_Product_Folder, file), os.path.join(OutputFolder, file))

    
    ### read attribute table
    finalcat_info = Dbf_To_Dataframe(Path_final_riv_ply)
    finalcat_info = finalcat_info.fillna(-9999)

    ### Obtain selected lake's attribute info
    (
        Selected_Non_ConnLakes,
        Selected_ConnLakes,
        Un_Selected_ConnLakes_info,
        Un_Selected_Non_ConnL_info,
    ) = Return_Selected_Lakes_Attribute_Table_And_Id(
        finalcat_info,
        Thres_Area_Conn_Lakes*1000*1000,
        Thres_Area_Non_Conn_Lakes*1000*1000,
        Selected_Lake_List_in,
    )

    ### Extract lake polygons
    if len(Selected_Non_ConnLakes) > 0:
        Selectfeatureattributes(
            processing,
            Input=Path_NonCon_Lake_ply,
            Output=os.path.join(OutputFolder, os.path.basename(Path_NonCon_Lake_ply)),
            Attri_NM="Hylak_id",
            Values=Selected_Non_ConnLakes,
        )
    if len(Selected_ConnLakes) > 0:
        Selectfeatureattributes(
            processing,
            Input=Path_Con_Lake_ply,
            Output=os.path.join(OutputFolder, os.path.basename(Path_Con_Lake_ply)),
            Attri_NM="Hylak_id",
            Values=Selected_ConnLakes,
        )

    ### create a copy of shapfiles in temp folder
    Path_Temp_final_rviply = os.path.join(
        tempfolder, "temp_finalriv_ply_selectlake.shp"
    )
    Path_Temp_final_rvi = os.path.join(tempfolder, "temp_finalriv_selectlake.shp")
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=Path_final_riv,
        FIELD=["SubId"],
        OUTPUT=Path_Temp_final_rvi,
    )
    qgis_vector_dissolve(
        processing,
        context,
        INPUT=Path_final_riv_ply,
        FIELD=["SubId"],
        OUTPUT=Path_Temp_final_rviply,
    )

    # change lake related attribute for un selected connected lake
    # catchment to -1.2345
    Remove_Unselected_Lake_Attribute_In_Finalcatinfo(
        Path_Temp_final_rviply, Selected_ConnLakes
    )
    # remove lake attributes
    Remove_Unselected_Lake_Attribute_In_Finalcatinfo(
        Path_Temp_final_rvi, Selected_ConnLakes
    )

    # Modify attribute table to merge un selected lake catchment if needed
    finalcat_info_temp = Dbf_To_Dataframe(Path_Temp_final_rviply)
    # change attributes for catchment due to remove of connected lakes
    mapoldnew_info = (
        Change_Attribute_Values_For_Catchments_Need_To_Be_Merged_By_Remove_CL(
            finalcat_info_temp, Un_Selected_ConnLakes_info
        )
    )
    # change attribute for catchment due to remove of non connected lakes
    mapoldnew_info = (
        Change_Attribute_Values_For_Catchments_Need_To_Be_Merged_By_Remove_NCL(
            mapoldnew_info, finalcat_info_temp, Un_Selected_Non_ConnL_info
        )
    )
    
    mapoldnew_info.loc[mapoldnew_info['HyLakeId'] <= 0,'HyLakeId'] = 0
    mapoldnew_info.loc[mapoldnew_info['HyLakeId'] <= 0,'Lake_Cat'] = 0
    mapoldnew_info.loc[mapoldnew_info['HyLakeId'] <= 0,'LakeVol'] = 0
    mapoldnew_info.loc[mapoldnew_info['HyLakeId'] <= 0,'LakeDepth'] = 0
    mapoldnew_info.loc[mapoldnew_info['HyLakeId'] <= 0,'LakeArea'] = 0
    mapoldnew_info.loc[mapoldnew_info['HyLakeId'] <= 0,'Laketype'] = 0
    
    # update topology for new attribute table
    UpdateTopology(mapoldnew_info, UpdateStreamorder=-1)
    mapoldnew_info = update_non_connected_catchment_info(mapoldnew_info)


    all_subids = finalcat_info_temp['SubId'].values
    
    copy_data_and_dissolve(all_subids,tempfolder,processing,Path_Temp_final_rviply,Path_Temp_final_rvi,
        mapoldnew_info,COLUMN_NAMES_CONSTANT_CLEAN,OutputFolder,Path_Catchment_Polygon,context,
        Path_final_riv_ply,Path_final_riv)
        
        
    # # copy new attribute table to shpfiles
    # Copy_Pddataframe_to_shpfile(
    #     Path_Temp_final_rviply,
    #     mapoldnew_info,
    #     link_col_nm_shp="SubId",
    #     link_col_nm_df="Old_SubId",
    #     UpdateColNM=["#"],
    # )
    # Copy_Pddataframe_to_shpfile(
    #     Path_Temp_final_rvi,
    #     mapoldnew_info,
    #     link_col_nm_shp="SubId",
    #     link_col_nm_df="Old_SubId",
    #     UpdateColNM=["#"],
    # )
    # 
    # # disslove line and polygon based on new subid
    # qgis_vector_dissolve(
    #     processing,
    #     context,
    #     INPUT=Path_Temp_final_rvi,
    #     FIELD=["SubId"],
    #     OUTPUT=os.path.join(OutputFolder, os.path.basename(Path_final_riv)),
    # )
    # qgis_vector_dissolve(
    #     processing,
    #     context,
    #     INPUT=Path_Temp_final_rviply,
    #     FIELD=["SubId"],
    #     OUTPUT=os.path.join(OutputFolder, os.path.basename(Path_final_riv_ply)),
    # )
    # 
    # # clean attribute table
    # Clean_Attribute_Name(
    #     os.path.join(OutputFolder, os.path.basename(Path_final_riv)),
    #     COLUMN_NAMES_CONSTANT,
    # )
    # Clean_Attribute_Name(
    #     os.path.join(OutputFolder, os.path.basename(Path_final_riv_ply)),
    #     COLUMN_NAMES_CONSTANT,
    # )

    return



import pandas as pd
import numpy as np
from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def export_files_to_output_folder(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    output_riv,
    output_cat,
    output_folder,
    obs_attributes,
):
    cat_riv_info = input_geo_names["cat_riv_info"]
    cat_ply_info = input_geo_names["cat_ply_info"]
    snapped_obs_points = input_geo_names["snapped_obs_points"]
    all_lakes = input_geo_names["all_lakes"]

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    processing.run(
        "native:dissolve",
        {
            "INPUT": os.path.join(grassdb, cat_riv_info + ".shp"),
            "FIELD": ["SubId"],
            "OUTPUT": os.path.join(grassdb, cat_riv_info + "_dis.shp"),
        },
        context=context,
    )
    processing.run(
        "native:dissolve",
        {
            "INPUT": os.path.join(grassdb, cat_ply_info + ".shp"),
            "FIELD": ["SubId"],
            "OUTPUT": os.path.join(grassdb, cat_ply_info + "_dis.shp"),
        },
        context=context,
    )

    subinfo = Dbf_To_Dataframe(os.path.join(grassdb, cat_ply_info + "_dis.shp"))

    layer_cat = QgsVectorLayer(os.path.join(grassdb, cat_ply_info + "_dis.shp"), "")
    # add attribute to layer
    layer_cat = qgis_vector_add_attributes(
        processing,
        context,
        INPUT_Layer=layer_cat,
        attribute_list=[
            QgsField("centroid_x", QVariant.Double),
            QgsField("centroid_y", QVariant.Double),
        ],
    )

    formular = "x(centroid(transform($geometry,'%s','%s')))" % (
        layer_cat.crs().authid(),
        "EPSG:4326",
    )

    layer_cat = qgis_vector_field_calculator(
        processing=processing,
        context=context,
        FORMULA=formular,
        FIELD_NAME="centroid_x",
        INPUT=layer_cat,
        OUTPUT="memory:",
    )["OUTPUT"]
        

    formular = "y(centroid(transform($geometry,'%s','%s')))" % (
        layer_cat.crs().authid(),
        "EPSG:4326",
    )

    layer_cat = qgis_vector_field_calculator(
        processing=processing,
        context=context,
        FORMULA=formular,
        FIELD_NAME="centroid_y",
        INPUT=layer_cat,
        OUTPUT="memory:",
    )["OUTPUT"]


    Selectfeatureattributes(
        processing,
        Input=layer_cat,
        Output=os.path.join(grassdb, output_cat + "withnegv.shp"),
        Attri_NM="SubId",
        Values=subinfo[subinfo["SubId"] > 0]["SubId"].values,
    )
    Selectfeatureattributes(
        processing,
        Input=os.path.join(grassdb, cat_riv_info + "_dis.shp"),
        Output=os.path.join(grassdb, output_riv + "withnegv.shp"),
        Attri_NM="SubId",
        Values=subinfo[subinfo["SubId"] > 0]["SubId"].values,
    )
    
    change_neg_value_to_null_in_attribute_table(
                 processing,
                 os.path.join(grassdb, output_cat + "withnegv.shp"),
                 os.path.join(output_folder, output_cat + ".shp")
    )
    change_neg_value_to_null_in_attribute_table(
                 processing,
                 os.path.join(grassdb, output_riv + "withnegv.shp"),
                 os.path.join(output_folder, output_riv + ".shp")
    )
    
    # Add_centroid_to_feature(
    #     os.path.join(output_folder, output_cat + ".shp"), "centroid_x", "centroid_y"
    # )

    subinfo = Dbf_To_Dataframe(os.path.join(output_folder, output_cat + ".shp"))

    if os.path.exists(os.path.join(grassdb, all_lakes + ".shp")):
        cl_lakeids = subinfo.loc[subinfo["Lake_Cat"] == 1]["HyLakeId"].values
        ncl_lakeids = subinfo.loc[subinfo["Lake_Cat"] == 2]["HyLakeId"].values

        if len(cl_lakeids) > 0:
            Selectfeatureattributes(
                processing,
                Input=os.path.join(grassdb, all_lakes + ".shp"),
                Output=os.path.join(output_folder, "sl_connected_lake.shp"),
                Attri_NM="Hylak_id",
                Values=cl_lakeids,
            )
        if len(ncl_lakeids) > 0:
            Selectfeatureattributes(
                processing,
                Input=os.path.join(grassdb, all_lakes + ".shp"),
                Output=os.path.join(output_folder, "sl_non_connected_lake.shp"),
                Attri_NM="Hylak_id",
                Values=ncl_lakeids,
            )

    if (
        os.path.exists(os.path.join(grassdb, snapped_obs_points + ".shp"))
        and len(subinfo.loc[subinfo["Has_POI"] > 0]["Has_POI"].values) > 0
    ):

        Selectfeatureattributes(
            processing,
            Input=os.path.join(grassdb, snapped_obs_points + ".shp"),
            Output=os.path.join(output_folder, "obs_gauges.shp"),
            Attri_NM=obs_attributes[1],
            Values=subinfo.loc[subinfo["Has_POI"] > 0]["Obs_NM"].values,
            Is_str = True,
        )

    Clean_Attribute_Name(
        os.path.join(output_folder, output_cat + ".shp"), COLUMN_NAMES_CONSTANT_CLEAN
    )
    Clean_Attribute_Name(
        os.path.join(output_folder, output_riv + ".shp"), COLUMN_NAMES_CONSTANT_CLEAN
    )

    return



from basinmaker.func.arcgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def define_project_extent_using_hybasin_ply(
    work_folder,
    grass_location,
    qgis_prefix_path,
    path_dem_in,
    buffer_distance,
    hybasin_ply,
    down_hybasin_id,
    up_hybasin_id,
    mask="MASK",
    dem="dem",
):

    """Define processing extent

    Function that used to define project processing spatial extent (PSE).
    The processing spatial extent is a region where Toolbox will work in. Toolbox
    will not process grids or features outside the processing spatial extent.
    Several options is available here. The PSE can be defined using Hybasin
    product and a hydrobasin ID. All subbasin drainage to that hydrobasin ID
    will be extracted. And the extent of the extracted polygon will be used as PSE

    Parameters
    ----------
    grassdb                           : path (required)
        It is a path to project grass database folder
    grass_location                    : string (required)
        It is a string of grass location name
    qgis_prefix_path                  : string (required)
        It is a string of qgis prefix path
    path_dem_in                      : string (required)
        It is the path to input dem
    buffer_distance                  : float (optional)
        It is a float number to increase the extent of the PSE
        obtained from Hydrobasins. It is needed when input DEM is not from
        HydroSHEDS. Then the extent of the watershed will be different
        with PSE defined by HydroBASINS.
    hybasin_ply                      : string (optional)
        It is a path to hydrobasin routing product, If it is provided, the
        PSE will be based on the OutHyID and OutHyID2 and
        this HydroBASINS routing product.
    down_hybasin_id                  : int (optional)
        It is a HydroBASINS subbasin ID, which should be the ID of the most
        downstream subbasin in the region of interest.
    up_hybasin_id                    : int (optional)
        It is a HydroBASINS subbasin ID, which should be the ID of the most
        upstream subbasin in the region of interest, normally do not needed.
    mask                             : string (optional)
        It is a output mask name, which will stored in grass_location in both
        vector and raster format
    dem                              : string (optional)
        It is a output dem raster name, which will be stored in grass_location

    Notes
    -------
    Outputs are following files

    MASK                   : raster
        it is a mask raster stored in grass database, which indicate
        the PSE. The grass database is located at
        os.path.join(grassdb, grass_location)
    dem                   : raster
        it is a dem raster stored in grass database, which is
        has the same extent with MASK. The grass database is located at
        os.path.join(grassdb, grass_location)

    Returns:
    -------
       None

    Examples
    -------
    """

    print("mask region:   using hybasin polygon ")

    # read and define arcgis work enviroments 
    arcpy.env.overwriteOutput = True
    arcpy.CheckOutExtension("Spatial")
    cellSize = float(arcpy.GetRasterProperties_management(path_dem_in, "CELLSIZEX").getOutput(0))
    SptailRef = arcpy.Describe(path_dem_in).spatialReference
    arcpy.env.XYTolerance = cellSize
    arcpy.arcpy.env.cellSize = cellSize    
    arcpy.env.outputCoordinateSystem = arcpy.SpatialReference(int(SptailRef.factoryCode)) ### WGS84
    if not os.path.exists(work_folder):
        os.makedirs(work_folder)
    arcpy.env.workspace = work_folder


    hyshdinfo = Dbf_To_Dataframe(hybasin_ply)
    hyshdinfo = pd.DataFrame.spatial.from_featureclass(hybasin_ply)
    
    routing_info = hyshdinfo[["HYBAS_ID", "NEXT_DOWN"]].astype("float").values

    # obtain sub id of subbasins between OutHyID and OutHyID2 in the routing
    # network
    HydroBasins = Return_SubIds_Between_Two_Subbasins_In_Rouing_Network(
        routing_info, down_hybasin_id, up_hybasin_id
    )

    hyshdinfo_select = hyshdinfo.loc[hyshdinfo['HYBAS_ID'].isin(HydroBasins)]
    hyshdinfo_select['MASK'] = 1
    
    hyshdinfo_select.spatial.to_featureclass(location=os.path.join(work_folder,mask + "_hy.shp"),overwrite=True,sanitize_columns=False)
    
    arcpy.Dissolve_management(mask + "_hy.shp", mask + "_1hy.shp", ['MASK'])
    
    arcpy.Buffer_analysis(mask + "_1hy.shp",  mask + "_2hy.shp", buffer_distance)
    
    arcpy.Project_management(
        mask + "_2hy.shp",
        mask + ".shp", 
        arcpy.SpatialReference(int(SptailRef.factoryCode)),
        )    

    dem_mask_raster = ExtractByMask(path_dem_in, mask + ".shp")
    dem_mask_raster.save(dem+'.tif')
    dem_mask_raster.save(mask+'.tif')

    return 



from basinmaker.utilities.utilities import Internal_Constant_Names


def divide_domain_into_sub_regions(
    input_geo_names,
    grassdb,
    grass_location,
    qgis_prefix_path,
    path_lakefile_in,
    lake_attributes,
    path_bkfwidthdepth="#",
    bkfwd_attributes="#",
    Min_Num_Domain=9,
    Max_Num_Domain=13,
    Initaial_Acc=5000,
    Delta_Acc=1000,
    CheckLakeArea=1,
    fdr_path="#",
    Acc_Thresthold_stream=500,
    max_memory=2048 * 3,
    Out_Sub_Reg_Folder="#",
    sub_reg_str_r="sub_reg_str_r",
    sub_reg_str_v="sub_reg_str_v",
    sub_reg_nfdr_grass="sub_reg_nfdr_grass",
    sub_reg_nfdr_arcgis="sub_reg_nfdr_arcgis",
    sub_reg_acc="sub_reg_acc",
    sub_reg_dem="sub_reg_dem",
    gis_platform="qgis",
):
    cat_add_lake = Internal_Constant_Names["cat_add_lake"]

    if gis_platform == "qgis":
        assert (
            grassdb != "#"
        ), "grass database folder is needed, when gis_platform = qgis"
        assert (
            grass_location != "#"
        ), "grass location name is needed, when gis_platform = qgis"
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.subreg.generatesubregionqgis import (
            Generatesubdomain,
            generatesubdomainmaskandinfo,
        )

        Generatesubdomain(
            input_geo_names=input_geo_names,
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            path_lakefile_in=path_lakefile_in,
            lake_attributes=lake_attributes,
            Min_Num_Domain=Min_Num_Domain,
            Max_Num_Domain=Max_Num_Domain,
            Initaial_Acc=Initaial_Acc,
            Delta_Acc=Delta_Acc,
            CheckLakeArea=CheckLakeArea,
            fdr_path=fdr_path,
            Acc_Thresthold_stream=Acc_Thresthold_stream,
            max_memory=max_memory,
            Out_Sub_Reg_Folder=Out_Sub_Reg_Folder,
            sub_reg_str_r=sub_reg_str_r,
            sub_reg_str_v=sub_reg_str_v,
            sub_reg_nfdr_grass=sub_reg_nfdr_grass,
            sub_reg_nfdr_arcgis=sub_reg_nfdr_arcgis,
            sub_reg_acc=sub_reg_acc,
            sub_reg_dem=sub_reg_dem,
            cat_add_lake=cat_add_lake,
        )

        input_geo_names["cat_add_lake"] = cat_add_lake
        generatesubdomainmaskandinfo(
            Out_Sub_Reg_Dem_Folder=Out_Sub_Reg_Folder,
            input_geo_names=input_geo_names,
            grassdb=grassdb,
            grass_location=grass_location,
            qgis_prefix_path=qgis_prefix_path,
            path_bkfwidthdepth=path_bkfwidthdepth,
            bkfwd_attributes=bkfwd_attributes,
        )


def combine_sub_region(
    path_sub_region_info,
    sub_region_outputfolder,
    outputfolder,
    is_final_result,
    qgis_prefix_path,
    path_subregion_inlet,
    gis_platform="qgis",
    start_sub_id = 0,
    k = 1,
    c = 1,
):
    if gis_platform == "qgis":
        assert (
            qgis_prefix_path != "#"
        ), "qgis prefix path is needed, when gis_platform = qgis"
        from basinmaker.subreg.generatesubregionqgis import Combine_Sub_Region_Results

        Combine_Sub_Region_Results(
            Sub_Region_info=path_sub_region_info,
            Sub_Region_OutputFolder=sub_region_outputfolder,
            OutputFolder=outputfolder,
            Is_Final_Result=is_final_result,
            qgis_prefix_path=qgis_prefix_path,
            subregion_inlet=path_subregion_inlet,
            start_sub_id = start_sub_id,
            k = k,
            c = c
        )



import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
from hymodin.raveninput import GenerateRavenInput

Path_final_hru_info =sys.argv[1]
Startyear=-1
EndYear=-1
CA_HYDAT='#'
Lake_As_Gauge=0
Model_Name = sys.argv[2]
OutputFolder = sys.argv[3]

if Startyear != '#':
    Startyear = int(Startyear)
    DownLoadObsData = True
else:
    Startyear = -1
    DownLoadObsData = False

if EndYear != '#':
    EndYear = int(EndYear)
else:
    EndYear = -1 
    
#Inmportance_order = Inmportance_order.split(";")

GenerateRavenInput(
    Path_final_hru_info=Path_final_hru_info,
    lenThres=1,
    iscalmanningn=1,
    Startyear=Startyear,
    EndYear=EndYear,
    CA_HYDAT=CA_HYDAT,
    WarmUp=0,
    Template_Folder="#",
    Lake_As_Gauge=Lake_As_Gauge,
    WriteObsrvt=False,
    DownLoadObsData=DownLoadObsData,
    Model_Name=Model_Name,
    Old_Product=False,
    SubBasinGroup_NM_Channel=["Allsubbasins"],
    SubBasinGroup_Length_Channel=[-1],
    SubBasinGroup_NM_Lake=["AllLakesubbasins"],
    SubBasinGroup_Area_Lake=[-1],
    OutputFolder=OutputFolder,
    Forcing_Input_File="#",
    aspect_from_gis = "arcgis",
)



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3
from basinmaker.preprocessing.preprocessingobs import preprocessing_obs_point


def add_obs_into_existing_watershed_delineation(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    path_obsfile_in,
    obs_attributes=[],
    search_radius=100,
    path_sub_reg_outlets_v="#",
    max_memroy=1024 * 4,
    pourpoints_add_obs="pourpoints_add_obs",
    snapped_obs_points="snapped_obs_points",
):

    fdr_arcgis = input_geo_names["fdr_arcgis"]
    fdr_grass = input_geo_names["fdr_grass"]
    str_r = input_geo_names["str_r"]
    str_v = input_geo_names["str_v"]
    acc = input_geo_names["acc"]
    cat_no_lake = input_geo_names["cat_no_lake"]
    mask = input_geo_names["mask"]
    dem = input_geo_names["dem"]
    pourpoints_with_lakes = input_geo_names["pourpoints_with_lakes"]
    lake_outflow_pourpoints = input_geo_names["lake_outflow_pourpoints"]
    cat_add_lake = input_geo_names["cat_add_lake"]

    # define internal file names
    obsname = Internal_Constant_Names["obs"]

    # prepropessing lakes inputs
    preprocessing_obs_point(
        mask=mask,
        path_obsin_in=path_obsfile_in,
        obs_attributes=obs_attributes,
        grassdb=grassdb,
        grass_location=grass_location,
        qgis_prefix_path=qgis_prefix_path,
        gis_platform="qgis",
        obsname=obsname + "t1",
    )

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    # obtain maximum current cat id

    catids, temp = generate_stats_list_from_grass_raster(
        grass, mode=1, input_a=pourpoints_with_lakes
    )

    maxcatid = max(catids)
    # snap obs points
    grass_raster_r_stream_snap(
        grass,
        input=obsname + "t1",
        output=obsname + "_snap",
        stream_rast=str_r,
        accumulation=acc,
        radius=search_radius,
        memory=max_memroy,
    )

    # obtain use Obs_ID as observation point raster value
    grass_raster_v_to_raster(
        grass,
        input=obsname + "_snap",
        output=snapped_obs_points,
        column="#",
        use="cat",
    )
    grass_raster_r_to_vect(
        grass,
        input=snapped_obs_points,
        output=snapped_obs_points,
        type="point",
        flags="v",
    )
    grass_raster_v_db_join(
        grass,
        map=snapped_obs_points,
        column="cat",
        other_table=obsname + "t1",
        other_column="cat",
    )
    exp = obs_attributes[0] + "n int"
    grass.run_command(
        "v.db.addcolumn", map=snapped_obs_points, columns=obs_attributes[0] + "n int"
    )
    grass.run_command(
        "v.db.update",
        map=snapped_obs_points,
        column=obs_attributes[0] + "n",
        qcol=obs_attributes[0] + " + " + str(int(maxcatid) + 1),
    )

    grass.run_command(
        "v.out.ogr",
        input=snapped_obs_points,
        output=os.path.join(grassdb, snapped_obs_points + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
    )

    grass_raster_v_to_raster(
        grass,
        input=snapped_obs_points,
        output=obsname + "1",
        column=obs_attributes[0] + "n",
        use="attr",
    )

    if path_sub_reg_outlets_v != "#":
        # unpack subregion outlet point
        grass_raster_v_unpack(
            grass, input=path_sub_reg_outlets_v, output="Sub_reg_outlets_pt"
        )
        # convert it to raster
        grass_raster_v_to_raster(
            grass,
            input="Sub_reg_outlets_pt",
            output="Sub_reg_outlets",
            column="reg_subid",
            use="attr",
        )
        # added into observation raster point
        exp = "%s = if(isnull(int(Sub_reg_outlets)),%s,Sub_reg_outlets)" % (
            obsname,
            obsname + "1",
        )
        grass_raster_r_mapcalc(
            grass,
            expression=exp,
        )

        grass_raster_setnull(
            grass,
            raster_nm=obsname,
            null_values=[-9999, 0],
            create_new_raster=False,
            new_raster_nm="#",
        )
    else:
        grass.run_command("g.copy", rast=(obsname + "1", obsname), overwrite=True)
    ####

    # remove obs point located in lake catchments
    if lake_outflow_pourpoints != "#":

        ##### obtain lake id and correspond catchment id
        lake_id, cat_id = generate_stats_list_from_grass_raster(
            grass,
            mode=2,
            input_a=lake_outflow_pourpoints,
            input_b=pourpoints_with_lakes,
        )

        lake_new_cat_ids = np.column_stack((lake_id, cat_id))
        grass.run_command("g.copy", rast=(obsname, obsname + "2"), overwrite=True)
        # remove obs that located within the lake catchments
        obsid, cat_add_lake_id = generate_stats_list_from_grass_raster(
            grass, mode=2, input_a=obsname, input_b=cat_add_lake
        )
        lakecat_obs = np.column_stack((cat_add_lake_id, obsid))
        obsinlake_mask = np.isin(lakecat_obs[:, 0], lake_new_cat_ids[:, 1])
        obsid_inlake = lakecat_obs[obsinlake_mask, 1]
        if len(obsid_inlake) > 0:
            grass.run_command(
                "r.null", map=obsname + "2", setnull=obsid_inlake, overwrite=True
            )

        # combine lake and obs pourpoints
        # combine obsoutlets and outlet from cat no lake
        exp = "'%s' =if(isnull(int(%s)),%s,%s)" % (
            pourpoints_add_obs,
            pourpoints_with_lakes,
            obsname + "2",
            pourpoints_with_lakes,
        )
        grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    else:
        # combine obsoutlets and outlet from cat no lake
        exp = "'%s' =if(isnull(int(%s)),%s,%s)" % (
            pourpoints_add_obs,
            pourpoints_with_lakes,
            obsname,
            pourpoints_with_lakes,
        )
        grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    return



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *


def delineate_watershed_no_lake_using_subregion_data(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    subreg_fdr_path,
    subreg_acc_path,
    subreg_str_v_path,
    subreg_str_r_path,
    fdr_arcgis,
    fdr_grass,
    str_r,
    str_v,
    acc,
    cat_no_lake,
    max_memroy,
):
    mask = input_geo_names["mask"]

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    write_grass_and_arcgis_fdr_rules(grassdb)

    # unpack subregion flow accumulation and flow direction data
    grass_raster_r_unpack(grass, input=subreg_fdr_path, output="dir_Arcgis1")
    grass_raster_r_unpack(grass, input=subreg_acc_path, output="grass_acc1")
    exp = "%s  = int(dir_Arcgis1)" % (fdr_arcgis)
    grass_raster_r_mapcalc(grass, expression=exp)
    # calcuate flow direction in grass format
    grass_raster_r_reclass(
        grass,
        input=fdr_arcgis,
        output=fdr_grass,
        rules=os.path.join(grassdb, "Arcgis2GrassDIR.txt"),
    )
    exp = "%s  = grass_acc1" % (acc)
    grass_raster_r_mapcalc(grass, expression=exp)

    # unpack  river network vector
    grass_raster_v_unpack(grass, input=subreg_str_v_path, output=str_v)
    # unpack  river network raster
    grass_raster_r_unpack(grass, input=subreg_str_r_path, output="str_grass_r1")
    # clip river network with mask
    exp = "%s = int(str_grass_r1)" % (str_r)
    grass_raster_r_mapcalc(grass, expression=exp)

    grass_raster_r_stream_basins(
        grass,
        direction=fdr_grass,
        stream=str_r,
        basins=cat_no_lake,
        memory=max_memroy,
    )

    PERMANENT.close()
    return



import pandas as pd
import numpy as np
from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import os
import sqlite3


def join_pandas_table_to_vector_attributes(
    grassdb,
    grass_location,
    qgis_prefix_path,
    vector_name,
    pd_table,
    column_types,
    columns_names,
):

    str = ",".join(column_types)
    WriteStringToFile(str, os.path.join(grassdb, "catinfo_riv.csvt"), "w")
    pd_table.loc[pd_table['Has_POI'] <= 0,'Has_POI'] = 0
    pd_table.loc[pd_table['Has_POI'] > 0,'Has_POI'] = 1
    pd_table.loc[pd_table['Lake_Cat'] <= 0,'Lake_Cat'] = 0
    pd_table.loc[pd_table['LakeArea'] <= 0,'LakeArea'] = 0
    pd_table.loc[pd_table['HyLakeId'] <= 0,'HyLakeId'] = 0
    pd_table.loc[pd_table['Laketype'] <= 0,'Laketype'] = 0
    pd_table.loc[pd_table['LakeVol'] <= 0,'LakeVol'] = 0
    pd_table.loc[pd_table['LakeDepth'] <= 0,'LakeDepth'] = 0
    pd_table.loc[pd_table['LakeArea'] <= 0,'LakeArea'] = 0
    pd_table.loc[pd_table['DA_Obs'] <= 0,'DA_Obs'] = 0
    pd_table.loc[pd_table['DA_error'] <= 0,'DA_error'] = 0
    
    pd_table.to_csv(os.path.join(grassdb, "catinfo_riv.csv"), index=None, header=True)

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )
    
#    pd_table.to_sql('catinfo_pd',con,if_exists='replace',index_label='ix')
    
    ### add catchment info to all river segment
    grass.run_command(
        "db.in.ogr",
        input=os.path.join(grassdb, "catinfo_riv.csv"),
        output="result_riv",
        overwrite=True,
    )
    grass.run_command(
        "v.db.join",
        map=vector_name,
        column="Gridcode",
        other_table="result_riv",
        other_column="SubId",
        overwrite=True,
    )

    grass.run_command(
        "v.db.update",
        map=vector_name,
        column="DA_error",
        qcol="DrainArea/1000/1000/DA_Obs",
        where="Has_POI > 0",
    )

    grass.run_command(
        "v.out.ogr",
        input=vector_name,
        output=os.path.join(grassdb, vector_name + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
        quiet="Ture",
    )



import numpy as np
import sys
import os
import csv
import tempfile
import copy
import pandas as pd
from arcgis.features import GeoAccessor, GeoSeriesAccessor
import arcpy
from arcpy import env
from arcpy.sa import *


import sys, os
sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

from basinmaker.func.arcgis import *
from basinmaker.func.pdtable import *
arcpy.env.overwriteOutput = True
arcpy.CheckOutExtension("Spatial")



def GenerateHRUS_arcgis(
    Path_Subbasin_Ply,
    Landuse_info,
    Soil_info,
    Veg_info,
    Inmportance_order,
    min_hru_area_pct_sub,
    Sub_Lake_ID="HyLakeId",
    Sub_ID="SubId",
    Path_Connect_Lake_ply="#",
    Path_Non_Connect_Lake_ply="#",
    Lake_Id="Hylak_id",
    Path_Landuse_Ply="#",
    Landuse_ID="Landuse_ID",
    Path_Soil_Ply="#",
    Soil_ID="Soil_ID",
    Path_Veg_Ply="#",
    Veg_ID="Veg_ID",
    Path_Other_Ply_1="#",
    Other_Ply_ID_1="O_ID_1",
    Path_Other_Ply_2="#",
    Other_Ply_ID_2="O_ID_2",
    DEM="#",
    Project_crs = '3573',
    OutputFolder="#",
):
    """Generate HRU polygons and their attributes needed by hydrological model

    Function that be used to overlay: subbasin polygon, lake polygon (optional)
    , Land use polygon (optional), soil type polygon(optional),
    vegetation polygon (optional), and two other user defined polygons
    (optional).
    Non-lake HRU polygons in a subbasin is defined by an unique
    combination of all user provided datasets.
    A lake HRU polygon is defined the same as the provided lake polygon.
    All value of landuse and Veg polygon covered by lake will
    be changed to 1, indicating it is a covered by lake.
    All value of the soil polygon covered by the lake will be change to
    the soil id of the polygon covered by the lake with largest area.

    Parameters
    ----------
    Path_Subbasin_Ply                 : string
        It is the path of the subbasin polygon, which is generated by
        toolbox. if not generated by toolbox, the attribute table should
        including following attribute.
        ##############Subbasin related attributes###########################
        SubID           - integer, The subbasin Id
        DowSubId        - integer, The downstream subbasin ID of this
                                   subbasin
        IsLake          - integer, If the subbasin is a lake / reservior
                                   subbasin. 1 yes, <0, no
        IsObs           - integer, If the subbasin contains a observation
                                   gauge. 1 yes, < 0 no.
        RivLength       - float,   The length of the river in current
                                   subbasin in m
        RivSlope        - float,   The slope of the river path in
                                   current subbasin, in m/m
        FloodP_n        - float,   Flood plain manning's coefficient, in -
        Ch_n            - float,   main channel manning's coefficient, in -
        BkfWidth        - float,   the bankfull width of the main channel
                                   in m
        BkfDepth        - float,   the bankfull depth of the main channel
                                   in m
        HyLakeId        - integer, the lake id
        LakeVol         - float,   the Volume of the lake in km3
        LakeDepth       - float,   the average depth of the lake m
        LakeArea        - float,   the area of the lake in m2
    Landuse_info                      : string
        Path to a csv file that contains landuse information, including
        following attributes:
        Landuse_ID (can be any string)  - integer, the landuse ID in the
                                                   landuse polygon
        LAND_USE_C                      - string,  the landuse class name
                                                   for each landuse Type
    Soil_info                        : string
        Path to a csv file that contains soil information, including
        following attributes:
        Soil_ID (can be any string)     - integer, the Soil ID in the
                                                   soil polygon
        SOIL_PROF                       - string,  the Soil profile name
                                                   for each soil type
    Veg_info                         : string
        Path to a csv file that contains vegetation information, including
        following attributes:
        Veg_ID (can be any string)      - integer, the vegetation ID in the
                                                   vegetation polygon
        VEG_C                           - string,  the vegetation class name
                                                   for each vegetation Type
    Sub_Lake_ID                      : string (optional)
        The column name of the lake id in the subbasin polygon
    Sub_ID                           : string (optional)
        The column name of the subbasin id in the subbasin polygon
    Path_Connect_Lake_ply            : string (Optional)
        Path to the connected lake's polygon
    Path_Non_Connect_Lake_ply        : string (Optional)
        Path to the non connected lake's polygon
    Lake_Id                          : string (Optional)
        The the column name in lake polygon indicate the lake ID.
    Path_Landuse_Ply                 : string (Optional)
        Path to the landuse polygon. when Path_Landuse_Ply is not
        provided. The Landuse ID in Landuse_info should be
        1: land, -1: lake
    Landuse_ID                       : string (Optional)
        the the column name in landuse polygon and Landuse_info csv
        indicate the landuse ID. when Path_Landuse_Ply is not
        provided. The Landuse ID should be
        1: land, -1: lake.
    Path_Soil_Ply                    : string (Optional)
        Path to the soil polygon. when soil polygon is not
        provided. The Soil ID in Soil_info should be the same
        as Landuse ID.
    Soil_ID                          : string (Optional)
        the the column name in soil polygon and soil_info csv
        indicate the soil ID. when soil polygon is not
        provided. The Soil ID in Soil_info should be the same
        as Landuse ID.
    Path_Veg_Ply                     : string (Optional)
        Path to the vegetation polygon. when Veg polygon is not
        provided. The Veg ID in Veg_info should be the same
        as Landuse ID.
    Veg_ID                           : string (Optional)
        the the column name in vegetation polygon and veg_info csv
        indicate the vegetation ID. when Veg polygon is not
        provided. The Veg ID in Veg_info should be the same
        as Landuse ID.
    Path_Other_Ply_1                 : string (Optional)
        Path to the other polygon that will be used to define HRU,
        such as elevation band, or aspect.
    Other_Ply_ID_1                   : string (Optional)
        the the column name in Other_Ply_1 polygon
        indicate the landuse ID.
    Path_Other_Ply_2                 : string (Optional)
        Path to the other polygon that will be used to define HRU,
        such as elevation band, or aspect.
    Other_Ply_ID_2                   : string (Optional)
        the the column name in Other_Ply_2 polygon
        indicate the landuse ID.
    DEM                              : string (optional)
        the path to a raster elevation dataset, that will be used to
        calcuate average apspect, elevation and slope within each HRU.
        if no data is provided, basin average value will be used for
        each HRU.
    Project_crs                      : string
        the EPSG code of a projected coodinate system that will be used to
        calcuate HRU area and slope.
    OutputFolder                     : string
        The path to the folder that will save output HRU polygon.

    Notes
    -------
    Following ouput files will be generated in "<OutputFolder>/"
    'finalcat_hru_info.shp'              - HRU polygon and it's attributes


    Returns:
    -------
       None

    Examples
    -------
    >>> from ToolboxClass import LRRT
    >>> import pandas as pd
    >>> DataFolder = "C:/Path_to_foldr_of_example_dataset_provided_in_Github_wiki/"
    >>> RTtool=LRRT()
    >>> RTtool.GenerateHRUS(OutputFolder = DataFolder,
                           Path_Subbasin_Ply = os.path.join(DataFolder,"finalcat_info.shp"),
                           Path_Connect_Lake_ply = os.path.join(DataFolder,'Con_Lake_Ply.shp'),
                           Path_Non_Connect_Lake_ply = os.path.join(DataFolder,'Non_Con_Lake_Ply.shp'),
                           Path_Landuse_Ply = os.path.join(DataFolder,'modislanduse_exp_lg_pre.shp'),
                           Landuse_ID = 'gridcode',
                           Path_Soil_Ply = os.path.join(DataFolder,'ca_all_slc_v3r2_exp_lg.shp'),
                           Soil_ID = 'POLY_ID',
                           Landuse_info=os.path.join(DataFolder,'landuse_info.csv'),
                           Soil_info=os.path.join(DataFolder,'soil_info.csv'),
                           Veg_info=os.path.join(DataFolder,'veg_info.csv'),
                           DEM = os.path.join(DataFolder,'na_dem_15s_1.tif')
                           )

    """

    if not os.path.exists(OutputFolder):
        os.makedirs(OutputFolder)

    tempfolder = os.path.join(
        tempfile.gettempdir(), "basinmaker_hru" + str(np.random.randint(1, 10000 + 1))
    )
    if not os.path.exists(tempfolder):
        os.makedirs(tempfolder)
    print(Project_crs[5:len(Project_crs)])
    prj_crs = int(Project_crs[5:len(Project_crs)])
    Merge_layer_list = []
    ### First overlay the subbasin layer with lake polygon, the new unique
    # id will be 'HRULake_ID'

    Sub_Lake_HRU_Layer, trg_crs, fieldnames_list = GeneratelandandlakeHRUS(
        OutputFolder,
        tempfolder,
        Path_Subbasin_ply=Path_Subbasin_Ply,
        Path_Connect_Lake_ply=Path_Connect_Lake_ply,
        Path_Non_Connect_Lake_ply=Path_Non_Connect_Lake_ply,
        Sub_ID=Sub_ID,
        Sub_Lake_ID=Sub_Lake_ID,
        Lake_Id=Lake_Id,
    )

    lakehruinfo = pd.DataFrame.spatial.from_featureclass(Sub_Lake_HRU_Layer)
    hru_lake_info = lakehruinfo.loc[lakehruinfo['HRU_IsLake'] > 0].copy()
    lakehruinfo_landhrus = lakehruinfo.loc[lakehruinfo['HRU_IsLake'] <= 0].copy()

    lakehruinfo_landhrus = lakehruinfo_landhrus.spatial.to_featureclass(
        location=os.path.join(tempfolder,'land_hrus.shp'),
        overwrite=True,sanitize_columns=False,
    )

    fieldnames_list.extend(
        [
            Landuse_ID,
            Soil_ID,
            Veg_ID,
            Other_Ply_ID_1,
            Other_Ply_ID_2,
            "LAND_USE_C",
            "VEG_C",
            "SOIL_PROF",
            "HRU_Slope",
            "HRU_Area",
            "HRU_Aspect",
            "SHAPE",
        ]
    )
    dissolve_filedname_list = ["HRULake_ID"]

    Merge_layer_list.append([os.path.join(tempfolder,'land_hrus.shp'),1])

    #### check which data will be inlucded to determine HRU
    if Path_Landuse_Ply != "#":
        layer_path = Reproj_Clip_Dissolve_Simplify_Polygon_arcgis(
            layer_path = Path_Landuse_Ply,
            Class_Col = Landuse_ID,
            tempfolder = tempfolder,
            mask_layer = os.path.join(tempfolder,'land_hrus.shp')
        )
        Merge_layer_list.append([layer_path,2])
        dissolve_filedname_list.append(Landuse_ID)

    if Path_Soil_Ply != "#":
        layer_path = Reproj_Clip_Dissolve_Simplify_Polygon_arcgis(
            layer_path = Path_Soil_Ply,
            Class_Col = Soil_ID,
            tempfolder = tempfolder,
            mask_layer = os.path.join(tempfolder,'land_hrus.shp')
        )
        Merge_layer_list.append([layer_path,2])
        dissolve_filedname_list.append(Soil_ID)

    if Path_Veg_Ply != "#":
        layer_path = Reproj_Clip_Dissolve_Simplify_Polygon_arcgis(
            layer_path = Path_Veg_Ply,
            Class_Col = Veg_ID,
            tempfolder = tempfolder,
            mask_layer = os.path.join(tempfolder,'land_hrus.shp')
        )
        Merge_layer_list.append([layer_path,2])
        dissolve_filedname_list.append(Veg_ID)

    if Path_Other_Ply_1 != "#":
        layer_path = Reproj_Clip_Dissolve_Simplify_Polygon_arcgis(
            layer_path = Path_Other_Ply_1,
            Class_Col = Other_Ply_ID_1,
            tempfolder = tempfolder,
            mask_layer = os.path.join(tempfolder,'land_hrus.shp')
        )
        Merge_layer_list.append([layer_path,2])
        dissolve_filedname_list.append(Other_Ply_ID_1)

    if Path_Other_Ply_2 != "#":
        layer_path = Reproj_Clip_Dissolve_Simplify_Polygon_arcgis(
            layer_path = Path_Other_Ply_2,
            Class_Col = Other_Ply_ID_2,
            tempfolder = tempfolder,
            mask_layer = os.path.join(tempfolder,'land_hrus.shp')
        )
        Merge_layer_list.append([layer_path,2])
        dissolve_filedname_list.append(Other_Ply_ID_2)

    fieldnames = np.array(fieldnames_list)

    print("begin union")
    #### uniion polygons in the Merge_layer_list
    print(Merge_layer_list)
    arcpy.Union_analysis(Merge_layer_list, os.path.join(tempfolder,'hrus_land_union.shp'))
    print("union done ")

    hru_land_info = pd.DataFrame.spatial.from_featureclass( os.path.join(tempfolder,'hrus_land_union.shp'))

    #####

    Landuse_info_data = pd.read_csv(Landuse_info)
    Soil_info_data = pd.read_csv(Soil_info)
    Veg_info_data = pd.read_csv(Veg_info)

    # landuse polygon is not provided,landused id the same as IS_lake
    if Path_Landuse_Ply == "#":
        hru_land_info[Landuse_ID] = -hru_land_info['HRU_IsLake']
    hru_lake_info[Landuse_ID] = -1
    # if soil is not provied, it the value,will be the same as land use
    if Path_Soil_Ply == "#":
        hru_land_info[Soil_ID] = hru_land_info[Landuse_ID]
    hru_lake_info[Soil_ID] = -1
    # if no vegetation polygon is provide vegetation, will be the same as landuse
    if Path_Veg_Ply == "#":
        hru_land_info[Veg_ID] = hru_land_info[Landuse_ID]
    hru_lake_info[Veg_ID] = -1
    if Path_Other_Ply_1 == "#":
        hru_land_info[Other_Ply_ID_1] = -hru_land_info['HRU_IsLake']
    hru_lake_info[Other_Ply_ID_1] = -1
    if Path_Other_Ply_2 == "#":
        hru_land_info[Other_Ply_ID_2] = -hru_land_info['HRU_IsLake']
    hru_lake_info[Other_Ply_ID_2] = -1

    hru_lake_info = clean_attribute_name_arcgis(hru_lake_info,fieldnames)
    hru_land_info = clean_attribute_name_arcgis(hru_land_info,fieldnames)


    hruinfo = pd.concat([hru_lake_info, hru_land_info], ignore_index=True)

    HRU_draf_final = Define_HRU_Attributes_arcgis(
        prj_crs = prj_crs,
        trg_crs = trg_crs,
        hruinfo = hruinfo,
        dissolve_filedname_list = dissolve_filedname_list,
        Sub_ID = Sub_ID,
        Landuse_ID = Landuse_ID,
        Soil_ID = Soil_ID,
        Veg_ID = Veg_ID,
        Other_Ply_ID_1 = Other_Ply_ID_1,
        Other_Ply_ID_2 = Other_Ply_ID_2,
        Landuse_info_data = Landuse_info_data,
        Soil_info_data = Soil_info_data,
        Veg_info_data = Veg_info_data,
        DEM = DEM,
        Path_Subbasin_Ply = Path_Subbasin_Ply,
        min_hru_area_pct_sub = min_hru_area_pct_sub,
        Inmportance_order = Inmportance_order,
        OutputFolder = OutputFolder,
        tempfolder = tempfolder,
    )

    COLUMN_NAMES_CONSTANT_HRU_extend = COLUMN_NAMES_CONSTANT_HRU.extend(
        [
            Landuse_ID,
            Soil_ID,
            Veg_ID,
            Other_Ply_ID_1,
            Other_Ply_ID_2,
        ]
    )
    HRU_draf_final = clean_attribute_name_arcgis(HRU_draf_final,COLUMN_NAMES_CONSTANT_HRU)
    HRU_draf_final.spatial.to_featureclass(location = os.path.join(OutputFolder,'finalcat_hru_info.shp'),overwrite=True,sanitize_columns=False)

def GeneratelandandlakeHRUS(
    OutputFolder,
    tempfolder,
    Path_Subbasin_ply,
    Path_Connect_Lake_ply="#",
    Path_Non_Connect_Lake_ply="#",
    Sub_ID="SubId",
    Sub_Lake_ID="HyLakeId",
    Lake_Id="Hylak_id",
):

    """Overlay subbasin polygon and lake polygons

    Function that will overlay subbasin polygon and lake polygon

    Parameters
    ----------
    processing                        : qgis object
    context                           : qgis object
    OutputFolder                     : string
        The path to the folder that will save output HRU polygon.

    Path_Subbasin_Ply                 : string
        It is the path of the subbasin polygon, which is generated by
        toolbox. if not generated by toolbox, the attribute table should
        including following attribute.
        ##############Subbasin related attributes###########################
        SubID           - integer, The subbasin Id
        DowSubId        - integer, The downstream subbasin ID of this
                                   subbasin
        IsLake          - integer, If the subbasin is a lake / reservior
                                   subbasin. 1 yes, <0, no
        IsObs           - integer, If the subbasin contains a observation
                                   gauge. 1 yes, < 0 no.
        RivLength       - float,   The length of the river in current
                                   subbasin in m
        RivSlope        - float,   The slope of the river path in
                                   current subbasin, in m/m
        FloodP_n        - float,   Flood plain manning's coefficient, in -
        Ch_n            - float,   main channel manning's coefficient, in -
        BkfWidth        - float,   the bankfull width of the main channel
                                   in m
        BkfDepth        - float,   the bankfull depth of the main channel
                                   in m
        HyLakeId        - integer, the lake id
        LakeVol         - float,   the Volume of the lake in km3
        LakeDepth       - float,   the average depth of the lake m
        LakeArea        - float,   the area of the lake in m2
    Path_Connect_Lake_ply            : string (Optional)
        Path to the connected lake's polygon
    Path_Non_Connect_Lake_ply        : string (Optional)
        Path to the non connected lake's polygon
    Sub_ID                           : string (optional)
        The column name of the subbasin id in the subbasin polygon
    Sub_Lake_ID                      : string (optional)
        The column name of the lake id in the subbasin polygon
    Lake_Id                          : string (Optional)
        The the column name in lake polygon indicate the lake ID.


    Notes
    -------
        None

    Returns:
    -------
        Sub_Lake_HRU['OUTPUT']                   : qgis object
            it is a polygon after overlay between subbasin polygon and
            lake polygon
        Sub_Lake_HRU['OUTPUT'].crs().authid()    : string
            it is a string indicate the geospatial reference used by SubBasin
            polygon
        ['HRULake_ID','HRU_IsLake',Sub_ID]       : list
            it is a string list
    """

    # Fix geometry errors in subbasin polygon
    arcpy.RepairGeometry_management(Path_Subbasin_ply)

    # Create a file name list that will be strored in output attribute table
    fieldnames_list = [
        "HRULake_ID",
        "HRU_IsLake",
        Lake_Id,
        Sub_ID,
        Sub_Lake_ID,
        "SHAPE"
    ]  ### attribubte name in the need to be saved
    fieldnames = np.array(fieldnames_list)

    # if no lake polygon is provided, use subId as HRULake_ID.
    if Path_Connect_Lake_ply == "#" and Path_Non_Connect_Lake_ply == "#":

        cat_info = pd.DataFrame.spatial.from_featureclass(Path_Subbasin_ply)
        cat_info['Hylak_id'] = -1
        cat_info['HRULake_ID'] = cat_info['SubId']
        cat_info['HRU_IsLake'] = -1

        # remove column not in fieldnames
        cat_info = clean_attribute_name_arcgis(cat_info,fieldnames)
        cat_info.spatial.to_featureclass(location=os.path.join(OutputFolder,'finalcat_hru_lake_info.shp'),overwrite=True,sanitize_columns=False)
        crs_id = arcpy.Describe(Path_Subbasin_ply).spatialReference.factoryCode
        return os.path.join(OutputFolder,'finalcat_hru_lake_info.shp'), crs_id, ["HRULake_ID", "HRU_IsLake", Sub_ID]

    # fix lake polygon  geometry
    if Path_Connect_Lake_ply != "#":
        arcpy.RepairGeometry_management(Path_Connect_Lake_ply)
    # fix lake polygon geometry
    if Path_Non_Connect_Lake_ply != "#":
        arcpy.RepairGeometry_management(Path_Non_Connect_Lake_ply)
    # Merge connected and non connected lake polygons first
    if Path_Connect_Lake_ply != "#" and Path_Non_Connect_Lake_ply != "#":
        arcpy.Merge_management([Path_Connect_Lake_ply, Path_Non_Connect_Lake_ply], os.path.join(tempfolder,'merged_lake_ply.shp'))
    elif Path_Connect_Lake_ply != "#" and Path_Non_Connect_Lake_ply == "#":
        arcpy.CopyFeatures_management(Path_Connect_Lake_ply, os.path.join(tempfolder,'merged_lake_ply.shp'))
    elif Path_Connect_Lake_ply == "#" and Path_Non_Connect_Lake_ply != "#":
        arcpy.CopyFeatures_management(Path_Non_Connect_Lake_ply, os.path.join(tempfolder,'merged_lake_ply.shp'))
    else:
        print("should never happened......")

    # union merged polygon and subbasin polygon
#    cat_info.spatial.to_featureclass(location=os.path.join(tempfolder,'cat_ply.shp'))
#    arcpy.RepairGeometry_management(os.path.join(tempfolder,'cat_ply.shp'))

    inFeatures = [[Path_Subbasin_ply, 1], [os.path.join(tempfolder,'merged_lake_ply.shp'), 2]]

    arcpy.Union_analysis(inFeatures, os.path.join(tempfolder,'cat_lake_union.shp'))

    arcpy.RepairGeometry_management(os.path.join(tempfolder,'cat_lake_union.shp'))
    sub_lake_info = pd.DataFrame.spatial.from_featureclass(location=os.path.join(tempfolder,'cat_lake_union.shp'))
    sub_lake_info['HRULake_ID'] = -9999
    sub_lake_info['HRU_IsLake'] = -9999

    crs_id = arcpy.Describe(os.path.join(tempfolder,'cat_lake_union.shp')).spatialReference.factoryCode

    sub_lake_info['HRU_ID_Temp'] = sub_lake_info['FID'] + 1

    sub_lake_info = Determine_Lake_HRU_Id(sub_lake_info)
    # copy determined lake hru id to vector
    sub_lake_info = clean_attribute_name_arcgis(sub_lake_info,fieldnames)
    save_modified_attributes_to_outputs(
        mapoldnew_info = sub_lake_info,
        tempfolder = tempfolder,
        OutputFolder = OutputFolder,
        cat_name = 'finalcat_hru_lake_info.shp',
        riv_name = '#',
        Path_final_riv = '#',
        dis_col_name='HRULake_ID'
    )

    return os.path.join(OutputFolder,'finalcat_hru_lake_info.shp'), crs_id, ["HRULake_ID", "HRU_IsLake", Sub_ID]


############


def Reproj_Clip_Dissolve_Simplify_Polygon_arcgis(
    layer_path, Class_Col, tempfolder,mask_layer
):
    """Preprocess user provided polygons

    Function that will reproject clip input polygon with subbasin polygon
    and will dissolve the input polygon based on their ID, such as landuse id
    or soil id.

    Parameters
    ----------
    processing                        : qgis object
    context                           : qgis object
    layer_path                        : string
        The path to a specific polygon, for example path to landuse layer
    Project_crs                       : string
        the EPSG code of a projected coodinate system that will be used to
        calcuate HRU area and slope.
    trg_crs                           : string
        the EPSG code of a  coodinate system that will be used to
        calcuate reproject input polygon
    Class_Col                         : string
        the column name in the input polygon (layer_path) that contains
        their ID, for example land use ID or soil ID.
    Layer_clip                        : qgis object
        A shpfile with extent of the watershed, will be used to clip input
        input polygon
    Notes
    -------
        # TODO: May be add some function to simplify the input polygons
                for example, remove the landuse type with small areas
                or merge small landuse polygon into the surrounding polygon

    Returns:
    -------
        layer_dis                  : qgis object
            it is a polygon after preprocess
    """
    arcpy.Project_management(
        layer_path,
        os.path.join(tempfolder,Class_Col+"_proj.shp"),
        arcpy.Describe(mask_layer).spatialReference,
        )
    arcpy.Clip_analysis(
        os.path.join(tempfolder,Class_Col+"_proj.shp"),
        mask_layer,
        os.path.join(tempfolder,Class_Col+"_clip.shp")
    )

    arcpy.Dissolve_management(
        os.path.join(tempfolder,Class_Col+"_clip.shp"),
        os.path.join(tempfolder,Class_Col+"_dislve.shp"),
        [Class_Col]
    )

    arcpy.RepairGeometry_management(os.path.join(tempfolder,Class_Col+"_dislve.shp"))

    arcpy.AddSpatialIndex_management(os.path.join(tempfolder,Class_Col+"_dislve.shp"))

    return os.path.join(tempfolder,Class_Col+"_dislve.shp")


def Union_Ply_Layers_And_Simplify(
    processing,
    context,
    Merge_layer_list,
    dissolve_filedname_list,
    fieldnames,
    OutputFolder,
):
    """Union input QGIS polygon layers

    Function will union polygon layers in Merge_layer_list
    dissove the union result based on field name in
    dissolve_filedname_list
    and remove all field names not in fieldnames

    Parameters
    ----------
    processing                        : qgis object
    context                           : qgis object
    Merge_layer_list                  : list
        a list contain polygons layers needs to be unioned
    dissolve_filedname_list           : list
        a list contain column name of ID in each polygon layer
        in Merge_layer_list
    fieldnames                        : list
        a list contain column names that are needed in the return
        QGIS polygon layers
    OutputFolder                      : String
        The path to a folder to save result during the processing

    Returns:
    -------
    mem_union_dis                  : qgis object
        it is a polygon object that generated by overlay all input
        layers
    """
    num = str(np.random.randint(1, 10000 + 1))
    ##union polygons
    if len(Merge_layer_list) == 1:
        mem_union = Merge_layer_list[0]
        mem_union_fix_ext = qgis_vector_fix_geometries(
            processing, context, INPUT=mem_union, OUTPUT="memory:"
        )["OUTPUT"]
    elif len(Merge_layer_list) > 1:
        for i in range(0, len(Merge_layer_list)):
            if i == 0:
                mem_union = Merge_layer_list[i]
                mem_union_fix_ext = qgis_vector_fix_geometries(
                    processing, context, INPUT=mem_union, OUTPUT="memory:"
                )["OUTPUT"]
                qgis_vector_create_spatial_index(processing, context, mem_union_fix_ext)
            else:
                mem_union_fix_temp = mem_union_fix_ext
                del mem_union_fix_ext
                mem_union = processing.run(
                    "saga:polygonunion",
                    {
                        "A": mem_union_fix_temp,
                        "B": Merge_layer_list[i],
                        "SPLIT": True,
                        "RESULT": "TEMPORARY_OUTPUT",
                    },
                    context=context,
                )["RESULT"]
                # mem_union = qgis_vector_union_two_layers(
                #     processing,
                #     context,
                #     mem_union_fix_temp,
                #     Merge_layer_list[i],
                #     "memory:",
                #     OVERLAY_FIELDS_PREFIX="",
                # )["OUTPUT"]
                mem_union_fix = qgis_vector_fix_geometries(
                    processing, context, INPUT=mem_union, OUTPUT="memory:"
                )["OUTPUT"]
                mem_union_fix_ext = mem_union_fix
                qgis_vector_create_spatial_index(processing, context, mem_union_fix_ext)
    else:
        print("No polygon needs to be overlaied.........should not happen ")

    ## remove non interested filed
    mem_union_fix_ext, temp_out_notused = Clean_Attribute_Name(
        mem_union_fix_ext, fieldnames, Input_Is_Feature_In_Mem=True
    )
    mem_union_dis = mem_union_fix_ext
    # qgis_vector_dissolve(
    #     processing,
    #     context,
    #     INPUT=mem_union_fix_ext,
    #     FIELD=dissolve_filedname_list,
    #     OUTPUT=os.path.join(tempfile.gettempdir(),num + "_dissolve_union.shp")
    # )
    #
    # mem_union_dis = qgis_vector_dissolve(
    #     processing,
    #     context,
    #     INPUT=os.path.join(tempfile.gettempdir(),num + "_dissolve_union.shp"),
    #     FIELD=dissolve_filedname_list,
    #     OUTPUT="memory:",
    # )["OUTPUT"]

    return mem_union_dis


def Define_HRU_Attributes_arcgis(
    prj_crs,
    trg_crs,
    hruinfo,
    dissolve_filedname_list,
    Sub_ID,
    Landuse_ID,
    Soil_ID,
    Veg_ID,
    Other_Ply_ID_1,
    Other_Ply_ID_2,
    Landuse_info_data,
    Soil_info_data,
    Veg_info_data,
    DEM,
    Path_Subbasin_Ply,
    Inmportance_order,
    min_hru_area_pct_sub,
    OutputFolder,
    tempfolder,
):

    """Generate attributes of each HRU

    Function will generate attributes that are needed by Raven and
    other hydrological models for each HRU

    Parameters
    ----------
    processing                        : qgis object
    context                           : qgis object
    Project_crs                       : string
        the EPSG code of a projected coodinate system that will be used to
        calcuate HRU area and slope.
    hru_layer                         : qgis object
        a polygon layer generated by overlay all input polygons
    dissolve_filedname_list           : list
        a list contain column name of ID in each polygon layer
        in Merge_layer_list
    Sub_ID                            : string
        The column name of the subbasin id in the subbasin polygon
    Landuse_ID                        : string
        the the column name in landuse polygon and Landuse_info csv
        indicate the landuse ID. when Path_Landuse_Ply is not
        provided. The Landuse ID should be
        1: land, -1: lake.
    Soil_ID                           : string
        the the column name in soil polygon and soil_info csv
        indicate the soil ID. when soil polygon is not
        provided. The Soil ID in Soil_info should be the same
        as Landuse ID.
    Veg_ID                            : string
        the the column name in vegetation polygon and veg_info csv
        indicate the vegetation ID. when Veg polygon is not
        provided. The Veg ID in Veg_info should be the same
        as Landuse ID.

    Landuse_info                      : Dataframe
        a dataframe that contains landuse information, including
        following attributes:
        Landuse_ID (can be any string)  - integer, the landuse ID in the
                                                   landuse polygon
        LAND_USE_C                      - string,  the landuse class name
                                                   for each landuse Type
    Soil_info                         : Dataframe
        a dataframe that contains soil information, including
        following attributes:
        Soil_ID (can be any string)     - integer, the Soil ID in the
                                                   soil polygon
        SOIL_PROF                       - string,  the Soil profile name
                                                   for each soil type
    Veg_info                          : Dataframe
        a dataframe file that contains vegetation information, including
        following attributes:
        Veg_ID (can be any string)      - integer, the vegetation ID in the
                                                   vegetation polygon
        VEG_C                           - string,  the vegetation class name
                                                   for each vegetation Type
    DEM                               : string (optional)
        the path to a raster elevation dataset, that will be used to
        calcuate average apspect, elevation and slope within each HRU.
        if no data is provided, basin average value will be used for
        each HRU.
    Path_Subbasin_Ply                 : string
        It is the path of the subbasin polygon, which is generated by
        toolbox. if not generated by toolbox, the attribute table should
        including following attribute.
        ##############Subbasin related attributes###########################
        SubID           - integer, The subbasin Id
        DowSubId        - integer, The downstream subbasin ID of this
                                   subbasin
        IsLake          - integer, If the subbasin is a lake / reservior
                                   subbasin. 1 yes, <0, no
        IsObs           - integer, If the subbasin contains a observation
                                   gauge. 1 yes, < 0 no.
        RivLength       - float,   The length of the river in current
                                   subbasin in m
        RivSlope        - float,   The slope of the river path in
                                   current subbasin, in m/m
        FloodP_n        - float,   Flood plain manning's coefficient, in -
        Ch_n            - float,   main channel manning's coefficient, in -
        BkfWidth        - float,   the bankfull width of the main channel
                                   in m
        BkfDepth        - float,   the bankfull depth of the main channel
                                   in m
        HyLakeId        - integer, the lake id
        LakeVol         - float,   the Volume of the lake in km3
        LakeDepth       - float,   the average depth of the lake m
        LakeArea        - float,   the area of the lake in m2
    OutputFolder                      : String
        The path to a folder to save result during the processing

    Returns:
    -------
    HRU_draf_final                  : qgis object
        it is a polygon object that generated by overlay all input
        layers and inlcude all needed attribue for hydrological model
        like RAVEN
    """
    num = str(np.random.randint(1, 10000 + 1))
    hruinfo["LAND_USE_C"] = '-9999'
    hruinfo["VEG_C"] = '-9999'
    hruinfo["SOIL_PROF"] = '-9999'
    hruinfo["HRU_CenX"] = -9999.9999
    hruinfo["HRU_CenY"] = -9999.9999
    hruinfo["HRU_ID_New"] = -9999
    hruinfo["HRU_Area"] = -9999.99
    hruinfo.spatial.to_featureclass(location=os.path.join(tempfolder,'hru_add_area.shp'),overwrite=True,sanitize_columns=False)

    arcpy.CalculateGeometryAttributes_management(
        os.path.join(tempfolder,'hru_add_area.shp'),
        [["HRU_Area", "AREA_GEODESIC"]],
        area_unit = 'SQUARE_METERS',
        coordinate_system = arcpy.SpatialReference(prj_crs)
    )
    ### calcuate area of each feature
    hruinfo_area = pd.DataFrame.spatial.from_featureclass(os.path.join(tempfolder,'hru_add_area.shp'))

    hruinfo_area['HRU_ID'] = hruinfo_area['FID'] + 1
    hruinfo_area["HRU_ID_New"] = hruinfo_area["FID"] + 1

    hruinfo_area_update_attribute = Determine_HRU_Attributes(
        hruinfo_area,
        Sub_ID,
        Landuse_ID,
        Soil_ID,
        Veg_ID,
        Other_Ply_ID_1,
        Other_Ply_ID_2,
        Landuse_info_data,
        Soil_info_data,
        Veg_info_data,
    )
    hruinfo_area_update_attribute['HRU_ID_New'] = hruinfo_area_update_attribute['HRU_ID_New'].astype(int)
    save_modified_attributes_to_outputs(
        mapoldnew_info = hruinfo_area_update_attribute,
        tempfolder = tempfolder,
        OutputFolder = tempfolder,
        cat_name = 'finalcat_hru_info.shp',
        riv_name = '#',
        Path_final_riv = '#',
        dis_col_name='HRU_ID_New'
    )

    arcpy.RepairGeometry_management(os.path.join(tempfolder,'finalcat_hru_info.shp'))

    arcpy.CalculateGeometryAttributes_management(
        os.path.join(tempfolder,'finalcat_hru_info.shp'),
        [["HRU_Area", "AREA_GEODESIC"]],
        area_unit = 'SQUARE_METERS',
        coordinate_system = arcpy.SpatialReference(prj_crs)
    )

    hruinfo_new = pd.DataFrame.spatial.from_featureclass(os.path.join(tempfolder,'finalcat_hru_info.shp'))
    hruinfo_simple = simplidfy_hrus(
        min_hru_pct_sub_area = min_hru_area_pct_sub,
        hruinfo = hruinfo_new,
        importance_order = Inmportance_order,
    )

    save_modified_attributes_to_outputs(
        mapoldnew_info = hruinfo_simple,
        tempfolder = tempfolder,
        OutputFolder = tempfolder,
        cat_name = 'hru_simple.shp',
        riv_name = '#',
        Path_final_riv = '#',
        dis_col_name='HRU_ID_New'
    )

    arcpy.CalculateGeometryAttributes_management(
        os.path.join(tempfolder,'hru_simple.shp'),
        [["HRU_CenX", "CENTROID_X"],
        ["HRU_CenY", "CENTROID_Y"]],
        coordinate_system = arcpy.SpatialReference(4326)
    )
    #
    arcpy.JoinField_management(
        os.path.join(tempfolder,'hru_simple.shp'),
        'SubId',
        os.path.join(Path_Subbasin_Ply),
        'SubId',
    )
    arcpy.CalculateGeometryAttributes_management(
        os.path.join(tempfolder,'hru_simple.shp'),
        [["HRU_Area", "AREA_GEODESIC"]],
        area_unit = 'SQUARE_METERS',
        coordinate_system = arcpy.SpatialReference(prj_crs)
    )

    if DEM != "#":

        arcpy.Project_management(
            os.path.join(tempfolder,'hru_simple.shp'),
            os.path.join(tempfolder,"hru_proj.shp"),
            arcpy.SpatialReference(int(prj_crs)),
            )

        extract_dem = ExtractByMask(DEM, os.path.join(tempfolder,'hru_simple.shp'))
        arcpy.ProjectRaster_management(
            extract_dem,
            os.path.join(tempfolder,"demproj.tif"),
            arcpy.SpatialReference(int(prj_crs)),
            "NEAREST"
        )
        Slopeout = Slope(extract_dem, "DEGREE", 0.3043)
        Slopeout.save(os.path.join(OutputFolder,'slope.tif'))
        Aspectout = Aspect(extract_dem)

        # Save the output
        Aspectout.save(os.path.join(OutputFolder,'aspect.tif'))
        table_zon_slope = ZonalStatisticsAsTable(
            os.path.join(tempfolder,"hru_proj.shp"),
            'HRU_ID_New',
            Slopeout,
            os.path.join(tempfolder,"slope_zonal.dbf"),
            "DATA",
            "MEAN",
        )
        table_zon_aspect = ZonalStatisticsAsTable(
            os.path.join(tempfolder,"hru_proj.shp"),
            'HRU_ID_New',
            Aspectout,
            os.path.join(tempfolder,"asp_zonal.dbf"),
            "DATA",
            "MEAN",
        )
        table_zon_elev = ZonalStatisticsAsTable(
            os.path.join(tempfolder,"hru_proj.shp"),
            'HRU_ID_New',
            os.path.join(tempfolder,"demproj.tif"),
            os.path.join(tempfolder,"elv_zonal.dbf"),
            "DATA",
            "MEAN",
        )
        hruinfo_add_slp_asp = pd.DataFrame.spatial.from_featureclass(os.path.join(tempfolder,'hru_simple.shp'))
        table_slp = Dbf_To_Dataframe(os.path.join(tempfolder,"slope_zonal.dbf"))
        table_asp = Dbf_To_Dataframe(os.path.join(tempfolder,"asp_zonal.dbf"))
        table_elv = Dbf_To_Dataframe(os.path.join(tempfolder,"elv_zonal.dbf"))

        table_slp['HRU_S_mean'] = table_slp['MEAN']
        table_slp = table_slp[['HRU_ID_New','HRU_S_mean']]
        table_asp['HRU_A_mean'] = table_asp['MEAN']
        table_asp = table_asp[['HRU_ID_New','HRU_A_mean']]
        table_elv['HRU_E_mean'] = table_elv['MEAN']
        table_elv = table_elv[['HRU_ID_New','HRU_E_mean']]


        hruinfo_add_slp_asp = pd.merge(hruinfo_add_slp_asp, table_slp, on='HRU_ID_New')
        hruinfo_add_slp_asp = pd.merge(hruinfo_add_slp_asp, table_asp, on='HRU_ID_New')
        hruinfo_add_slp_asp = pd.merge(hruinfo_add_slp_asp, table_elv, on='HRU_ID_New')
        hruinfo_add_slp_asp['HRU_ID'] = hruinfo_add_slp_asp['FID'] + 1
    else:
        arcpy.AddMessage(os.path.join(tempfolder,'hru_simple.shp'))
        hruinfo_add_slp_asp = pd.DataFrame.spatial.from_featureclass(os.path.join(tempfolder,'hru_simple.shp'))
        hruinfo_add_slp_asp['HRU_ID'] = hruinfo_add_slp_asp['FID'] + 1
        hruinfo_add_slp_asp['HRU_S_mean'] = hruinfo_add_slp_asp['BasSlope']
        hruinfo_add_slp_asp['HRU_A_mean'] = hruinfo_add_slp_asp['BasAspect']
        hruinfo_add_slp_asp['HRU_E_mean'] = hruinfo_add_slp_asp['MeanElev']

    return hruinfo_add_slp_asp



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3


def add_gauge_attributes(
    grassdb,
    grass_location,
    qgis_prefix_path,
    catinfo,
    input_geo_names,
    obs_attributes=[],
):

    outlet_pt_info = input_geo_names["outlet_pt_info"]
    snapped_obs_points = input_geo_names["snapped_obs_points"]
    obsname = "obs"  # input_geo_names["obsname"]

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    grass.run_command(
        "v.what.rast",
        map=outlet_pt_info,
        raster=obsname,
        column="obsid_pour",
    )

    grass.run_command(
        "v.out.ogr",
        input=outlet_pt_info,
        output=os.path.join(grassdb, outlet_pt_info + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
        quiet="Ture",
    )

    ### read catchment
    sqlstat = "SELECT SubId,obsid_pour FROM %s" % (outlet_pt_info,)
    outletinfo = pd.read_sql_query(sqlstat, con)
    outletinfo = outletinfo.fillna(-9999)
    outletinfo = outletinfo.loc[outletinfo["obsid_pour"] > 0]

    sqlstat = "SELECT  %s,%s,%s,%s,%s FROM %s" % (
        obs_attributes[0],
        obs_attributes[1],
        obs_attributes[2],
        obs_attributes[3],
        obs_attributes[0] + "n",
        snapped_obs_points,
    )
    gaugeinfo = pd.read_sql_query(sqlstat, con)
    gaugeinfo = gaugeinfo.fillna(-9999)

    for i in range(0, len(outletinfo)):
        catid = outletinfo["SubId"].values[i]
        obsid = outletinfo["obsid_pour"].values[i]
        catrow = catinfo["SubId"] == catid
        subgaugeinfo = gaugeinfo.loc[gaugeinfo[obs_attributes[0] + "n"] == obsid]

        catinfo.loc[catrow, "Has_POI"] = obsid
        if obsid > 79000:
            continue

        if len(subgaugeinfo) > 0:
            catinfo.loc[catrow, "Has_POI"] = subgaugeinfo[obs_attributes[0]].values[0]
            catinfo.loc[catrow, "DA_Obs"] = subgaugeinfo[obs_attributes[2]].values[0]
            catinfo.loc[catrow, "Obs_NM"] = subgaugeinfo[obs_attributes[1]].values[0]
            catinfo.loc[catrow, "SRC_obs"] = subgaugeinfo[obs_attributes[3]].values[0]

    PERMANENT.close()
    return catinfo



from basinmaker.func.grassgis import *
from basinmaker.func.qgis import *
from basinmaker.func.pdtable import *
from basinmaker.func.rarray import *
from basinmaker.utilities.utilities import *
import sqlite3


def calculate_basic_attributes(
    grassdb,
    grass_location,
    qgis_prefix_path,
    input_geo_names,
    projection,
    catinfo,
    cat_ply_info,
    cat_riv_info,
    outlet_pt_info,
):

    catchments = input_geo_names["catchment_without_merging_lakes"]
    river_r = input_geo_names["river_without_merging_lakes"]
    river_v = input_geo_names["str_v"]
    dem = input_geo_names["dem"]
    fdr = input_geo_names["nfdr_grass"]
    acc = input_geo_names["acc"]
    cat_use_default_acc = input_geo_names["cat_use_default_acc"]
    problem_seg = input_geo_names["problem_seg"]
    nfdr_arcgis = input_geo_names["nfdr_arcgis"]

    import grass.script as grass
    import grass.script.setup as gsetup
    from grass.pygrass.modules import Module
    from grass.pygrass.modules.shortcuts import general as g
    from grass.pygrass.modules.shortcuts import raster as r
    from grass.script import array as garray
    from grass.script import core as gcore
    from grass_session import Session

    QgsApplication.setPrefixPath(qgis_prefix_path, True)
    Qgs = QgsApplication([], False)
    Qgs.initQgis()
    from processing.core.Processing import Processing
    from processing.tools import dataobjects
    from qgis import processing

    feedback = QgsProcessingFeedback()
    Processing.initialize()
    QgsApplication.processingRegistry().addProvider(QgsNativeAlgorithms())
    context = dataobjects.createContext()
    context.setInvalidGeometryCheck(QgsFeatureRequest.GeometryNoCheck)

    os.environ.update(
        dict(GRASS_COMPRESS_NULLS="1", GRASS_COMPRESSOR="ZSTD", GRASS_VERBOSE="1")
    )
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")

    con = sqlite3.connect(
        os.path.join(grassdb, grass_location, "PERMANENT", "sqlite", "sqlite.db")
    )

    # create a catchment vector and overlay with str v
    exp = "%s = int(%s)" % (catchments, catchments)
    grass.run_command("r.mapcalc", expression=exp, overwrite=True)

    grass.run_command(
        "r.to.vect",
        input=catchments,
        output=cat_ply_info + "t1",
        type="area",
        overwrite=True,
    )
    grass.run_command(
        "v.db.addcolumn", map=cat_ply_info + "t1", columns="GC_str VARCHAR(40)"
    )
    grass.run_command(
        "v.db.addcolumn", map=cat_ply_info + "t1", columns="Area_m double"
    )
    grass.run_command(
        "v.db.update", map=cat_ply_info + "t1", column="GC_str", qcol="value"
    )
    # dissolve based on gridcode
    grass.run_command(
        "v.dissolve",
        input=cat_ply_info + "t1",
        column="GC_str",
        output=cat_ply_info,
        overwrite=True,
    )

    grass.run_command("v.db.addcolumn", map=cat_ply_info, columns="Gridcode INT")
    grass.run_command("v.db.update", map=cat_ply_info, column="Gridcode", qcol="GC_str")

    ## obtain a stream vector, segmentation based on new catchment polygon
    grass.run_command(
        "v.overlay",
        ainput=river_v,
        alayer=2,
        atype="line",
        binput=cat_ply_info,
        operator="and",
        output=cat_riv_info + "t1",
        overwrite=True,
    )
    grass.run_command(
        "v.out.ogr",
        input=cat_riv_info + "t1",
        output=os.path.join(grassdb, cat_riv_info + "t1" + ".shp"),
        format="ESRI_Shapefile",
        overwrite=True,
    )
    processing.run(
        "gdal:dissolve",
        {
            "INPUT": os.path.join(grassdb, cat_riv_info + "t1" + ".shp"),
            "FIELD": "b_GC_str",
            "OUTPUT": os.path.join(grassdb, cat_riv_info + ".shp"),
        },
    )
    grass.run_command(
        "v.import",
        input=os.path.join(grassdb, cat_riv_info + ".shp"),
        output=cat_riv_info,
        overwrite=True,
    )

    grass.run_command("v.db.addcolumn", map=cat_riv_info, columns="Gridcode INT")
    grass.run_command("v.db.addcolumn", map=cat_riv_info, columns="Length_m double")
    grass.run_command(
        "v.db.update", map=cat_riv_info, column="Gridcode", qcol="b_GC_str"
    )
    grass.run_command("v.db.dropcolumn", map=cat_riv_info, columns=["b_GC_str"])

    grass.run_command(
        "r.out.gdal",
        input=dem,
        output=os.path.join(grassdb, "dem_par.tif"),
        format="GTiff",
        overwrite=True,
    )

    PERMANENT.close()

    os.system(
        "gdalwarp "
        + '"'
        + os.path.join(grassdb, "dem_par.tif")
        + '"'
        + "    "
        + '"'
        + os.path.join(grassdb, "dem_proj.tif")
        + '"'
        + " -t_srs  "
        + '"'
        + projection
        + '"'
    )

    project = Session()
    project.open(
        gisdb=grassdb, location=grass_location + "_proj", create_opts=projection
    )

    grass.run_command(
        "r.import",
        input=os.path.join(grassdb, "dem_proj.tif"),
        output="dem_proj",
        overwrite=True,
    )
    grass.run_command("g.region", raster="dem_proj")
    grass.run_command(
        "v.proj",
        location=grass_location,
        mapset="PERMANENT",
        input=cat_riv_info,
        overwrite=True,
    )
    grass.run_command(
        "v.proj",
        location=grass_location,
        mapset="PERMANENT",
        input=cat_ply_info,
        overwrite=True,
    )

    grass.run_command(
        "v.to.db",
        map=cat_ply_info,
        option="area",
        columns="Area_m",
        units="meters",
        overwrite=True,
    )
    grass.run_command(
        "v.to.db",
        map=cat_riv_info,
        option="length",
        columns="Length_m",
        units="meters",
        overwrite=True,
    )

    grass.run_command(
        "r.slope.aspect",
        elevation="dem_proj",
        slope="slope",
        aspect="aspect",
        precision="DCELL",
        overwrite=True,
    )

    ### calcuate averaged slope and aspect of each subbasin
    grass.run_command(
        "v.rast.stats",
        map=cat_ply_info,
        raster="slope",
        column_prefix="s",
        method="average",
    )
    grass.run_command(
        "v.rast.stats",
        map=cat_ply_info,
        raster="aspect",
        column_prefix="a",
        method="average",
    )
    
    grass.run_command(
        "v.rast.stats",
        map=cat_ply_info,
        raster="dem_proj",
        column_prefix="d",
        method="average",
    )
    
    ### calcuate minimum and maximum dem along the channel
    grass.run_command(
        "v.rast.stats",
        map=cat_riv_info,
        raster="dem_proj",
        column_prefix="d",
        method=["minimum", "maximum"],
    )

    project.close()

    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")
    grass.run_command("g.region", raster="dem")
    grass.run_command(
        "v.proj",
        location=grass_location + "_proj",
        mapset="PERMANENT",
        input=cat_riv_info,
        overwrite=True,
    )        
    grass.run_command(
        "v.proj",
        location=grass_location + "_proj",
        mapset="PERMANENT",
        input=cat_ply_info,
        overwrite=True,
    )
    
    ### calcuate averaged DEM in each subbasin
    # grass.run_command(
    #     "v.rast.stats",
    #     map=cat_ply_info,
    #     raster="dem",
    #     column_prefix="d",
    #     method="average",
    # )
    # 
    # ### calcuate minimum and maximum dem along the channel
    # grass.run_command(
    #     "v.rast.stats",
    #     map=cat_riv_info,
    #     raster="dem",
    #     column_prefix="d",
    #     method=["minimum", "maximum"],
    # )
    
    ## get routing structure

    exp = "%s = int(%s)" % (fdr, fdr)
    grass.run_command(
        "r.mapcalc",
        expression=exp,
        overwrite=True,
    )

    grass.run_command(
        "r.accumulate",
        direction=fdr,
        accumulation="acc_grass_CatOL",
        overwrite=True,
    )

    ##### obtain catment id overlaied with problem seg
    prom_seg_id, cat_pro_id = generate_stats_list_from_grass_raster(
        grass, mode=2, input_a=problem_seg, input_b=catchments
    )
    cat_pro_id = np.unique(cat_pro_id)
    grass.run_command("g.copy", rast=(catchments, cat_use_default_acc), overwrite=True)
    if len(cat_pro_id) > 0:
        grass.run_command(
            "r.null", map=cat_use_default_acc, setnull=cat_pro_id, overwrite=True
        )
    else:
        cat_pro_id = []

    exp = "acc_grass_CatOL2 = if(isnull(%s),%s,%s)" % (
        cat_use_default_acc,
        "acc_grass_CatOL",
        acc,
    )
    grass.run_command(
        "r.mapcalc",
        expression=exp,
        overwrite=True,
    )

    routing_temp = generate_routing_info_of_catchments(
        grass,
        con,
        cat=catchments,
        acc="acc_grass_CatOL2",
        Name="Final",
        str=river_r,
        garray = garray,
    )

    routing_temp = generate_routing_info_of_catchments(
        grass,
        con,
        cat=river_r,
        acc="acc_grass_CatOL2",
        Name="Friv",
        str=river_r,
        garray = garray,
    )

    cat_array = garray.array(mapname=catchments)
    nfdr_arcgis_array = garray.array(mapname=nfdr_arcgis)
    cat_outlet_array = garray.array(mapname="Final_OL")
    ncols = int(cat_array.shape[1])
    nrows = int(cat_array.shape[0])

    grass.run_command("g.copy", vector=("Final_OL_v", outlet_pt_info), overwrite=True)
    PERMANENT.close()
    
    
    ## add coordinates to outlet point in wgs84 system
    project_wgs84 = Session()
    project_wgs84.open(
        gisdb=grassdb, location=grass_location + "_wgs84", create_opts='EPSG:4326'
    )
    grass.run_command(
        "v.proj",
        location=grass_location,
        mapset="PERMANENT",
        input=outlet_pt_info,
        overwrite=True,
    )
    grass.run_command(
        "v.to.db",
        map = outlet_pt_info,
        type = 'point',
        option = 'coor',
        columns = ['outletLng','outletLat'],
        overwrite=True,
    )  
    project_wgs84.close()

    ## import updated outlet points after add coordinates in wgs84 system
    PERMANENT = Session()
    PERMANENT.open(gisdb=grassdb, location=grass_location, create_opts="")
    grass.run_command("g.region", raster="dem")
    grass.run_command(
        "v.proj",
        location=grass_location + "_wgs84",
        mapset="PERMANENT",
        input=outlet_pt_info,
        overwrite=True,
    )   
 
    ### update dataframe
    ###
    sqlstat = "SELECT Gridcode, Length_m, d_minimum, d_maximum FROM %s" % (cat_riv_info)
    leninfo = pd.read_sql_query(sqlstat, con)
    ### read catchment
    sqlstat = "SELECT Gridcode, Area_m,d_average,s_average,a_average FROM %s" % (
        cat_ply_info
    )
    areainfo = pd.read_sql_query(sqlstat, con)

    ### read catchment
    sqlstat = "SELECT SubId, DowSubId,ILSubIdmax,ILSubIdmin,outletLat, outletLng FROM %s" % (outlet_pt_info)
    outletinfo = pd.read_sql_query(sqlstat, con)
    outletinfo = outletinfo.fillna(-9999)
    outletinfo = outletinfo.loc[outletinfo['SubId'] >= 0]
    ### read catchment
    sqlstat = "SELECT SubId, DSubId_str FROM %s" % ('Friv_OL_v')
    outlet_riv_info = pd.read_sql_query(sqlstat, con)
    outlet_riv_info = outlet_riv_info.fillna(-9999)


    leninfo = leninfo.astype(float).fillna(-9999)
    areainfo = areainfo.astype(float).fillna(-9999)

    for i in range(0, len(outletinfo)):
        catid = outletinfo["SubId"].values[i]
        DownSubID_cat = outletinfo["DowSubId"].values[i]
        outlet_lat = outletinfo["outletLat"].values[i]
        outlet_lon = outletinfo["outletLng"].values[i]
        
        catinfo.loc[i, "SubId"] = catid
        catinfo.loc[i, "outletLat"] = outlet_lat
        catinfo.loc[i, "outletLng"] = outlet_lon

        # load routing info based on str
        outlet_riv_info_i = outlet_riv_info.loc[outlet_riv_info['SubId'] == catid]

        # check if riv outlet exist
        if len(outlet_riv_info_i) < 1:
            DownSubID = DownSubID_cat
            DownSubID_riv = -9999
        else:
           # get down sub id based in river network
           DownSubID_riv = outlet_riv_info_i['DSubId_str'].values[0]

           # may need to modify if downsub id from river != down subid from cat
           if DownSubID_riv != DownSubID_cat:

               # check if DownSubID_cat drainage to DSubId_str
               Down_id_of_downriv_info = outletinfo.loc[outletinfo['SubId'] == DownSubID_riv]
               if len(Down_id_of_downriv_info) > 0:
                   # get down subid of down subid of river
                   DSubId_DSubId_str = Down_id_of_downriv_info['DowSubId'].values[0]
                   # if down subid of down sub id of river  = down sub id from cat
                   if catid == 10762:
                       print(DSubId_DSubId_str,DownSubID_cat)

                   if DSubId_DSubId_str == DownSubID_cat:
                       DownSubID = DownSubID_riv
                   else:
                       DownSubID = DownSubID_cat
               else:
                   DownSubID = DownSubID_cat
           else:
               DownSubID = DownSubID_cat

        if catid in cat_pro_id or DownSubID == catid or DownSubID == -9999:
            downsubid_array = return_subid_of_next_down_stream_grids(
                cat_array, catid, nfdr_arcgis_array, cat_outlet_array, ncols, nrows
            )
            #            print(catid,DownSubID,downsubid_array)
            if downsubid_array > 0:
                DownSubID = downsubid_array

        ### change the downsub id to -1 for watershed outlet
        if len(outletinfo.loc[outletinfo["SubId"] == DownSubID]) < 1:
            catinfo.loc[i, "DowSubId"] = -1
        elif catid == DownSubID:
            catinfo.loc[i, "DowSubId"] = -1
        else:
            catinfo.loc[i, "DowSubId"] = DownSubID

        catarea = np.unique(
            areainfo.loc[areainfo["Gridcode"] == catid]["Area_m"].values
        )  #'Area_m'
        catslope = np.unique(
            areainfo.loc[areainfo["Gridcode"] == catid]["s_average"].values
        )
        catelev = np.unique(
            areainfo.loc[areainfo["Gridcode"] == catid]["d_average"].values
        )
        cataspect = np.unique(
            areainfo.loc[areainfo["Gridcode"] == catid]["a_average"].values
        )
        if len(catarea) == 1:
            catinfo.loc[i, "BasArea"] = catarea
            catinfo.loc[i, "BasSlope"] = catslope
            catinfo.loc[i, "BasAspect"] = cataspect
            catinfo.loc[i, "MeanElev"] = catelev
        else:
            print(
                "Warning  basin area of stream  ",
                catid,
                "   need check   ",
                len(catarea),
            )
            catinfo.loc[i, "BasArea"] = -9999
            catinfo.loc[i, "BasSlope"] = -9999
            catinfo.loc[i, "BasAspect"] = -9999
            catinfo.loc[i, "MeanElev"] = -9999

        ### add river parameters
        rivlen = np.unique(
            leninfo.loc[leninfo["Gridcode"] == catid]["Length_m"].values
        )  #'Area_m'
        dmaxelev = np.unique(
            leninfo.loc[leninfo["Gridcode"] == catid]["d_maximum"].values
        )
        dminelev = np.unique(
            leninfo.loc[leninfo["Gridcode"] == catid]["d_minimum"].values
        )
        if len(rivlen) > 0:
            catinfo.loc[i, "RivLength"] = rivlen[0]
            maxdem = dmaxelev[0]
            mindem = dminelev[0]
            catinfo.loc[i, "Min_DEM"] = mindem
            catinfo.loc[i, "Max_DEM"] = maxdem
            if rivlen[0] >= 0:
                if max(0, float((maxdem - mindem)) / float(rivlen[0])) == 0:
                    slope_rch = -9999
                else:
                    slope_rch = max(
                        0, float((maxdem - mindem)) / float(rivlen[0])
                    )

                slope_rch = max(slope_rch,min_riv_slope)  
                slope_rch = min(slope_rch,max_riv_slope)
                  
                catinfo.loc[i, "RivSlope"] = slope_rch
                
            else:
                catinfo.loc[i, "RivSlope"] = -9999
        else:
            catinfo.loc[i, "RivLength"] = -9999
            catinfo.loc[i, "RivSlope"] = -9999
            catinfo.loc[i, "FloodP_n"] = -9999
            catinfo.loc[i, "Min_DEM"] = -9999
            catinfo.loc[i, "Max_DEM"] = -9999

    PERMANENT.close()
    return catinfo
